# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import base64
import ipaddress
import json
import logging
import os
import pathlib
import ssl
from typing import Optional, Union

import anyio

from ._async_utils import NamedTemporaryFile, check_output
from ._config import KubeConfigSet
from ._types import PathType

logger = logging.getLogger(__name__)


class KubeAuth:
    """Load kubernetes auth from kubeconfig, service account, or url."""

    def __init__(
        self,
        kubeconfig: Optional[PathType] = None,
        url: Optional[str] = None,
        serviceaccount: Optional[str] = None,
        namespace: Optional[str] = None,
        context: Optional[str] = None,
    ) -> None:
        self.server: str = ""
        self.client_cert_file: Optional[PathType] = None
        self.client_key_file: Optional[PathType] = None
        self.server_ca_file: Optional[PathType] = None
        self.token: Optional[str] = None
        self._namespace: Optional[str] = namespace
        self.active_context: str = ""
        self.kubeconfig: KubeConfigSet
        self.tls_server_name: Optional[str] = None
        self._url: str = ""
        if url:
            self._url = url
        self._insecure_skip_tls_verify: bool = False
        self._use_context: Optional[str] = context
        self._context: dict = {}
        self._cluster: dict = {}
        self._user: dict = {}
        self._serviceaccount: str = (
            serviceaccount
            if serviceaccount is not None
            else "/var/run/secrets/kubernetes.io/serviceaccount"
        )
        self._kubeconfig_path_or_dict: Union[dict, PathType]
        if kubeconfig:
            self._kubeconfig_path_or_dict = kubeconfig
        else:
            self._kubeconfig_path_or_dict = os.environ.get(
                "KUBECONFIG", "~/.kube/config"
            )

        self.__auth_lock: anyio.Lock = anyio.Lock()

    def __await__(self):
        async def f():
            await self.reauthenticate()
            return self

        return f().__await__()

    async def reauthenticate(self) -> None:
        """Reauthenticate with the server."""
        async with self.__auth_lock:
            if self._url:
                logger.debug("URL specified manually")
                self.server = self._url
            else:
                if self._kubeconfig_path_or_dict is not False:
                    await self._load_kubeconfig()
                if self._serviceaccount and not self.server:
                    await self._load_service_account()
            if not self.server:
                raise ValueError("Unable to find valid credentials")

    async def ssl_context(self):
        if self._insecure_skip_tls_verify:
            return False
        async with self.__auth_lock:
            if (
                not self.client_key_file
                and not self.client_cert_file
                and not self.server_ca_file
            ):
                # If no cert information is provided, fall back to default verification
                return True
            sslcontext = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            if self.client_key_file and self.client_cert_file:
                sslcontext.load_cert_chain(
                    certfile=self.client_cert_file,
                    keyfile=self.client_key_file,
                    password=None,
                )
            if self.server_ca_file:
                sslcontext.load_verify_locations(cafile=self.server_ca_file)
            return sslcontext

    @property
    def namespace(self) -> str:
        return self._namespace if self._namespace else "default"

    @namespace.setter
    def namespace(self, value: str):
        self._namespace = value

    async def _load_kubeconfig(self) -> None:
        """Load kubernetes auth from kubeconfig."""
        logger.debug("Loading kubeconfig")
        if isinstance(self._kubeconfig_path_or_dict, str) or isinstance(
            self._kubeconfig_path_or_dict, pathlib.Path
        ):
            try:
                if os.name != "nt":
                    self.kubeconfig = await KubeConfigSet(
                        *str(self._kubeconfig_path_or_dict).split(":")
                    )
                else:
                    # Windows doesn't support multiple configs in a path
                    self.kubeconfig = await KubeConfigSet(self._kubeconfig_path_or_dict)
            except ValueError:
                return
        else:
            self.kubeconfig = await KubeConfigSet(self._kubeconfig_path_or_dict)
        if self._use_context:
            try:
                self._context = self.kubeconfig.get_context(self._use_context)
                self.active_context = self._use_context
            except ValueError as e:
                raise ValueError(f"No such context {self._use_context}") from e
        elif self.kubeconfig.current_context:
            self._context = self.kubeconfig.get_context(self.kubeconfig.current_context)
            self.active_context = self.kubeconfig.current_context
        else:
            self._context = self.kubeconfig.contexts[0]["context"]
            self.active_context = self.kubeconfig.contexts[0]["name"]

        # Load configuration options from the context
        if self._namespace is None:
            self._namespace = self.kubeconfig.current_namespace

        # If no cluster is found in the context, assume it's a service account
        if not self._context["cluster"]:
            return

        self._cluster = self.kubeconfig.get_cluster(self._context["cluster"])
        self._user = self.kubeconfig.get_user(self._context["user"])
        self.server = self._cluster["server"]

        if (
            "insecure-skip-tls-verify" in self._cluster
            and self._cluster["insecure-skip-tls-verify"]
        ):
            self._insecure_skip_tls_verify = True

        if "tls-server-name" in self._cluster:
            self.tls_server_name = self._cluster["tls-server-name"]

        if "exec" in self._user:
            if (
                self._user["exec"]["apiVersion"]
                == "client.authentication.k8s.io/v1alpha1"
            ):
                raise ValueError(
                    "client.authentication.k8s.io/v1alpha1 is not supported for exec auth"
                )
            command = self._user["exec"]["command"]
            args = self._user["exec"].get("args") or []
            env = os.environ.copy()
            env.update(
                **{e["name"]: e["value"] for e in self._user["exec"].get("env") or []}
            )
            data = json.loads(await check_output(command, *args, env=env))["status"]
            if "token" in data:
                self._user["token"] = data["token"]
            elif "clientCertificateData" in data and "clientKeyData" in data:
                self._user["client-certificate-data"] = data["clientCertificateData"]
                self._user["client-key-data"] = data["clientKeyData"]
            else:
                raise KeyError(f"Did not find credentials in {command} output.")

        if "client-key" in self._user:
            client_key_path = anyio.Path(self._user["client-key"])
            if await client_key_path.exists():
                self.client_key_file = self._user["client-key"]
            else:
                self.client_key_file = (
                    anyio.Path(self.kubeconfig.get_path(self._use_context)).parent
                    / client_key_path
                )
        if "client-key-data" in self._user:
            async with NamedTemporaryFile(delete=False) as key_file:
                if "-----" in self._user["client-key-data"]:
                    key_data = self._user["client-key-data"].encode()
                else:
                    key_data = base64.b64decode(self._user["client-key-data"])
                await key_file.write_bytes(key_data)
                self.client_key_file = str(key_file)
        if "client-certificate" in self._user:
            client_cert_path = anyio.Path(self._user["client-certificate"])
            if await client_cert_path.exists():
                self.client_cert_file = self._user["client-certificate"]
            else:
                self.client_cert_file = (
                    anyio.Path(self.kubeconfig.get_path(self._use_context)).parent
                    / client_cert_path
                )
        if "client-certificate-data" in self._user:
            async with NamedTemporaryFile(delete=False) as cert_file:
                if "-----" in self._user["client-certificate-data"]:
                    cert_data = self._user["client-certificate-data"].encode()
                else:
                    cert_data = base64.b64decode(self._user["client-certificate-data"])
                await cert_file.write_bytes(cert_data)
                self.client_cert_file = str(cert_file)
        if "certificate-authority" in self._cluster:
            server_ca_path = anyio.Path(self._cluster["certificate-authority"])
            if await server_ca_path.exists():
                self.server_ca_file = self._cluster["certificate-authority"]
            else:
                self.server_ca_file = (
                    anyio.Path(self.kubeconfig.get_path(self._use_context)).parent
                    / server_ca_path
                )
        if "certificate-authority-data" in self._cluster:
            async with NamedTemporaryFile(delete=False) as ca_file:
                if "-----" in self._cluster["certificate-authority-data"]:
                    ca_data = self._cluster["certificate-authority-data"].encode()
                else:
                    ca_data = base64.b64decode(
                        self._cluster["certificate-authority-data"]
                    )

                await ca_file.write_bytes(ca_data)
                self.server_ca_file = str(ca_file)
        if "token" in self._user:
            self.token = self._user["token"]
        if "username" in self._user or "password" in self._user:
            raise ValueError(
                "username/password authentication was removed in Kubernetes 1.19, "
                "kr8s doesn't not support this Kubernetes version"
            )
        if "auth-provider" in self._user:
            if p := self._user["auth-provider"]["name"] != "oidc":
                raise ValueError(
                    f"auth-provider {p} was deprecated in Kubernetes 1.21 "
                    "and is not supported by kr8s"
                )
            # TODO: Handle refreshing OIDC token if missing or expired
            self.token = self._user["auth-provider"]["config"]["id-token"]

    async def _load_service_account(self) -> None:
        """Load credentials from service account."""
        logger.debug("Loading service account")
        self._serviceaccount = os.path.expanduser(self._serviceaccount)
        if not os.path.isdir(self._serviceaccount):
            return
        self.server = self._format_server_address(
            os.environ["KUBERNETES_SERVICE_HOST"],
            os.environ["KUBERNETES_SERVICE_PORT"],
        )
        async with await anyio.open_file(
            os.path.join(self._serviceaccount, "token")
        ) as f:
            self.token = await f.read()
        self.server_ca_file = os.path.join(self._serviceaccount, "ca.crt")
        if self.namespace is None:
            async with await anyio.open_file(
                os.path.join(self._serviceaccount, "namespace")
            ) as f:
                self.namespace = await f.read()

    @staticmethod
    def _format_server_address(host, port):
        try:
            ipaddress.IPv6Address(host)
            return f"https://[{host}]:{port}"
        except ipaddress.AddressValueError:
            return f"https://{host}:{port}"



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
from typing import List

import anyio
import rich.table
import typer
from rich import box
from rich.console import Console
from rich.live import Live

import kr8s
from kr8s._async_utils import anext
from kr8s.asyncio.objects import Table

TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

console = Console()


async def draw_table(kind, response, resource_names):
    table = rich.table.Table(box=box.SIMPLE)
    table.add_column("Namespace", style="magenta", no_wrap=True)

    for column in response.column_definitions:
        if column["priority"] == 0:
            kwargs = {}
            if column["name"] == "Name":
                kwargs = {"style": "cyan", "no_wrap": True}
            table.add_column(column["name"], **kwargs)

    for row in response.rows:
        if not resource_names or row["object"]["metadata"]["name"] == resource_names[0]:
            r = [
                str(row)
                for row, column in zip(row["cells"], response.column_definitions)
                if column["priority"] == 0
            ]
            table.add_row(row["object"]["metadata"]["namespace"], *r)

    if not table.rows:
        if resource_names:
            console.print(f"No resources found with name {resource_names[0]}.")
        else:
            console.print(f"No {kind} resources found in current namespace.")
        return None
    return table


async def get_resources(resources, label_selector, field_selector):
    data = {}
    for kind in resources:
        data[kind] = await anext(
            kr8s.asyncio.get(
                kind,
                label_selector=label_selector,
                field_selector=field_selector,
                as_object=Table,
            )
        )
    return data


async def get(
    resources: List[str] = typer.Argument(..., help="TYPE[.VERSION][.GROUP]"),
    all_namespaces: bool = typer.Option(
        False,
        "-A",
        "--all-namespaces",
        help="If present, list the requested object(s) across all namespaces. "
        "Namespace in current context is ignored even if specified with --namespace.",
    ),
    namespace: str = typer.Option(
        None,
        "-n",
        "--namespace",
    ),
    label_selector: str = typer.Option(
        "",
        "-l",
        "--selector",
        help="Selector (label query) to filter on, supports '=', '==', and '!='. "
        "(e.g. -l key1=value1,key2=value2). "
        "Matching objects must satisfy all of the specified label constraints.",
    ),
    field_selector: str = typer.Option(
        "",
        "--field-selector",
        help="Selector (field query) to filter on, supports '=', '==', and '!='. "
        "(e.g. --field-selector key1=value1,key2=value2). "
        "The server only supports a limited number of field queries per type. ",
    ),
    show_kind: bool = typer.Option(
        False,
        "--show-kind",
        help="If present, list the resource type for the requested object(s).",
    ),
    show_labels: bool = typer.Option(
        False,
        "--show-labels",
        help="When printing, show all labels as the last column "
        "(default hide labels column).",
    ),
    label_columns: List[str] = typer.Option(
        [],
        "-L",
        "--label-columns",
        help="Accepts a comma separated list of labels that are going to be presented as columns. "
        "Names are case-sensitive. "
        "You can also use multiple flag options like -L label1 -L label2...",
    ),
    watch: bool = typer.Option(
        False,
        "-w",
        "--watch",
        help="After listing/getting the requested object, watch for changes.",
    ),
):
    """Display one or many resources.

    Prints a table of the most important information about the specified resources. You can filter the list using a
    label selector and the --selector flag. If the desired resource type is namespaced you will only see results
    in your current namespace unless you pass --all-namespaces.

    By specifying the output as 'template' and providing a Jinja2 template as the value of the --template flag, you
    can filter the attributes of the fetched resources.

    Use "kubectl-ng api-resources" for a complete list of supported resources.
    """
    resource_names = []
    # Support kubectl-ng get pod,svc
    if len(resources) == 1:
        resources = resources[0].split(",")
        if len(resources) > 1 and watch:
            raise typer.BadParameter("you may only specify a single resource type")
    # Support kubectl-ng get pod [name]
    elif len(resources) == 2:
        resources, resource_names = ([r] for r in resources)
    else:
        raise typer.BadParameter(
            f"Format '{' '.join(resources)}' not currently supported."
        )
    kubernetes = await kr8s.asyncio.api()
    if namespace:
        kubernetes.namespace = namespace
    if all_namespaces:
        kubernetes.namespace = kr8s.ALL

    data = await get_resources(resources, label_selector, field_selector)

    if not watch:
        for kind, response in data.items():
            table = await draw_table(kind, response, resource_names)
            if table:
                console.print(table)
    else:
        kind = list(data)[0]
        response = data[kind]
        table = await draw_table(kind, response, resource_names)
        with Live(table, console=console, auto_refresh=False) as live:
            while True:
                await anyio.sleep(5)
                data = await get_resources(resources, label_selector, field_selector)
                # TODO handle changes in resources
                for kind, response in data.items():
                    table = await draw_table(kind, response, resource_names)
                live.update(table)
                live.refresh()



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
"""Utilities for working with Kubernetes data structures."""
from __future__ import annotations

import re
from typing import Any, Callable


def list_dict_unpack(
    input_list: list[dict], key: str = "key", value: str = "value"
) -> dict:
    """Convert a list of dictionaries to a single dictionary.

    Args:
        input_list: The list of dictionaries to convert to a single dictionary.
        key: The key to use for the new dictionary's keys. Defaults to "key".
        value: The key to use for the new dictionary's values. Defaults to "value".

    Returns:
        A dictionary with the keys and values from the input list.
    """
    return {i[key]: i[value] for i in input_list}


def dict_list_pack(
    input_dict: dict, key: str = "key", value: str = "value"
) -> list[dict]:
    """Convert a dictionary to a list of dictionaries.

    Args:
        input_dict: The dictionary to convert to a list of dictionaries.
        key: The key to use for the input dictionary's keys. Defaults to "key".
        value: The key to use for the input dictionary's values. Defaults to "value".

    Returns:
        A list of dictionaries with the keys and values from the input dictionary.
    """
    return [{key: k, value: v} for k, v in input_dict.items()]


def dot_to_nested_dict(dot_notated_key: str, value: Any) -> dict:
    """Convert a dot notated key to a nested dictionary.

    Args:
        dot_notated_key: The dot notated key to convert to a nested dictionary.
        value: The value to assign to the innermost key.

    Returns:
        A nested dictionary with the innermost key being the value of the
        dot notated key.
    """
    keys = dot_notated_key.split(".")
    nested_dict: dict = {}
    for key in reversed(keys):
        if not nested_dict:
            nested_dict[key] = value
        else:
            nested_dict = {key: nested_dict}
    return nested_dict


def dict_to_selector(selector_dict: dict) -> str:
    """Convert a dictionary to a Kubernetes selector.

    Args:
        selector_dict: The dictionary to convert to a Kubernetes selector.

    Returns:
        A Kubernetes selector string.
    """
    return ",".join(f"{k}={v}" for k, v in selector_dict.items())


def xdict(*in_dict, **kwargs):
    """Dictionary constructor that ignores None values.

    Args:
        in_dict: A dict to convert. Only one is allowed.
        **kwargs: Keyword arguments to be converted to a dict.

    Returns:
        A dict with None values removed.

    Raises:
        ValueError
            If more than one positional argument is passed, or if both a positional
            argument and keyword arguments are passed.

    Examples:
        >>> xdict(foo="bar", baz=None)
        {"foo": "bar"}

        >>> xdict({"foo": "bar", "baz": None})
        {"foo": "bar"}
    """
    if len(in_dict) > 1:
        raise ValueError(
            f"xdict expected at most 1 positional argument, got {len(in_dict)}"
        )
    if len(in_dict) == 1 and kwargs:
        raise ValueError(
            "xdict expected at most 1 positional argument, or multiple keyword arguments, got both"
        )
    if len(in_dict) == 1:
        [kwargs] = in_dict
    return {k: v for k, v in kwargs.items() if v is not None}


def sort_versions(
    versions: list[Any], key: Callable = lambda x: x, reverse: bool = False
) -> list[Any]:
    """Sort a list of Kubernetes versions by priority.

    Follows the spcification
    https://kubernetes.io/docs/tasks/extend-kubernetes/custom-resources/custom-resource-definition-versioning/#version-priority

    Args:
        versions: A list of Kubernetes versions to sort.
        key: A function to extract the version string from each element in the list.
            Defaults to the identity function
        reverse: If True, sort in descending order. Defaults to False

    Returns:
        A list of Kubernetes versions sorted by priority.

    Examples:
        >>> sort_versions(["v1", "v2", "v2beta1"])
        ["v2", "v1", "v2beta1"]

        >>> sort_versions(["v1beta2", "foo1", "foo10", "v1"])
        ["v1", "v1beta2", "foo1", "foo10"]
    """
    pattern = r"^v\d+((alpha|beta)\d+)?$"
    stable = []
    alphas = []
    betas = []
    others = []
    for version in versions:
        if re.match(pattern, key(version)) is not None:
            if "alpha" in key(version):
                alphas.append(version)
            elif "beta" in key(version):
                betas.append(version)
            else:
                stable.append(version)
        else:
            others.append(version)

    stable = sorted(stable, key=lambda v: int(key(v)[1:]), reverse=True)
    betas = sorted(
        betas, key=lambda v: tuple(map(int, key(v)[1:].split("beta"))), reverse=True
    )
    alphas = sorted(
        alphas, key=lambda v: tuple(map(int, key(v)[1:].split("alpha"))), reverse=True
    )
    others = sorted(others, key=lambda v: key(v))

    output = stable + betas + alphas + others
    if reverse:
        output.reverse()
    return output



# SPDX-FileCopyrightText: Copyright (c) 2024-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License

import typer
from rich.console import Console

from kr8s.asyncio.objects import Node

console = Console()


# Missing Options
# TODO --dry-run='none'
# TODO -l, --selector=''


async def cordon(
    node: str = typer.Argument(..., help="NODE"),
):
    """Mark node as unschedulable.

    Examples:
        # Mark node "foo" as unschedulable
        kubectl-ng cordon foo
    """
    nodes = [await Node.get(node)]
    for node_instance in nodes:
        await node_instance.cordon()
        console.print(f"node/{node_instance.name} cordoned")


async def uncordon(
    node: str = typer.Argument(..., help="NODE"),
):
    """Mark node as schedulable.

    Examples:
        # Mark node "foo" as schedulable
        kubectl-ng uncordon foo
    """
    nodes = [await Node.get(node)]
    for node_instance in nodes:
        await node_instance.uncordon()
        console.print(f"node/{node_instance.name} uncordoned")



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import base64
import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path

import pytest
import yaml

import kr8s
from kr8s._async_utils import anext
from kr8s._auth import KubeAuth
from kr8s._config import KubeConfig
from kr8s._testutils import set_env

HERE = Path(__file__).parent.resolve()


@pytest.fixture
async def kubeconfig_with_exec(k8s_cluster):
    # Open kubeconfig and extract the certificates
    kubeconfig = yaml.safe_load(k8s_cluster.kubeconfig_path.read_text())
    user = kubeconfig["users"][0]["user"]

    # Create exec command that produces the certificates
    with set_env(
        KUBE_CLIENT_CERTIFICATE_DATA=user["client-certificate-data"],
        KUBE_CLIENT_KEY_DATA=user["client-key-data"],
    ):
        kubeconfig["users"][0]["user"] = {
            "exec": {
                "apiVersion": "client.authentication.k8s.io/v1beta1",
                "command": sys.executable,
                "args": [str(HERE / "scripts" / "envexec.py")],
            }
        }
        with tempfile.NamedTemporaryFile() as f:
            f.write(yaml.safe_dump(kubeconfig).encode())
            f.flush()
            yield f.name


@pytest.fixture
async def kubeconfig_with_token(k8s_cluster, k8s_token):
    # Open kubeconfig and extract the certificates
    kubeconfig = yaml.safe_load(k8s_cluster.kubeconfig_path.read_text())
    kubeconfig["users"][0]["user"] = {"token": k8s_token}
    with tempfile.NamedTemporaryFile() as f:
        f.write(yaml.safe_dump(kubeconfig).encode())
        f.flush()
        yield f.name


@pytest.fixture
async def kubeconfig_with_decoded_certs(k8s_cluster):
    kubeconfig = yaml.safe_load(k8s_cluster.kubeconfig_path.read_text())
    kubeconfig["clusters"][0]["cluster"]["certificate-authority-data"] = (
        base64.b64decode(
            kubeconfig["clusters"][0]["cluster"]["certificate-authority-data"]
        )
    ).decode()
    kubeconfig["users"][0]["user"]["client-certificate-data"] = (
        base64.b64decode(kubeconfig["users"][0]["user"]["client-certificate-data"])
    ).decode()
    kubeconfig["users"][0]["user"]["client-key-data"] = (
        base64.b64decode(kubeconfig["users"][0]["user"]["client-key-data"])
    ).decode()
    with tempfile.NamedTemporaryFile() as f:
        f.write(yaml.safe_dump(kubeconfig).encode())
        f.flush()
        yield f.name


@pytest.fixture
async def kubeconfig_with_line_breaks_in_certs(k8s_cluster):
    def insert_every(instring: str, substring: str, interval: int) -> str:
        """Insert a substring every interval characters in instring.

        Example:
            >>> insert_every("abcdefghi", ".", 3)
            "abc.def.ghi"
        """
        return substring.join(
            instring[i : i + interval] for i in range(0, len(instring), interval)
        )

    kubeconfig = yaml.safe_load(k8s_cluster.kubeconfig_path.read_text())
    kubeconfig["clusters"][0]["cluster"]["certificate-authority-data"] = insert_every(
        kubeconfig["clusters"][0]["cluster"]["certificate-authority-data"], "\n", 64
    )
    kubeconfig["users"][0]["user"]["client-certificate-data"] = insert_every(
        kubeconfig["users"][0]["user"]["client-certificate-data"], "\n", 64
    )
    kubeconfig["users"][0]["user"]["client-key-data"] = insert_every(
        kubeconfig["users"][0]["user"]["client-key-data"], "\n", 64
    )
    with tempfile.NamedTemporaryFile() as f:
        f.write(yaml.safe_dump(kubeconfig).encode())
        f.flush()
        yield f.name


@pytest.fixture
async def kubeconfig_with_second_context(k8s_cluster):
    # Open kubeconfig and extract the certificates
    kubeconfig = yaml.safe_load(k8s_cluster.kubeconfig_path.read_text())
    kubeconfig["contexts"].append(
        {"context": kubeconfig["contexts"][0]["context"], "name": "foo-context"}
    )
    with tempfile.NamedTemporaryFile() as f:
        f.write(yaml.safe_dump(kubeconfig).encode())
        f.flush()
        yield f.name, kubeconfig["contexts"][1]["name"]


@pytest.fixture
async def kubeconfig_with_certs_on_disk(k8s_cluster):
    @contextmanager
    def f(absolute=True):
        # Open kubeconfig and dump certs to disk, then write new kubeconfig with paths to certs
        kubeconfig = yaml.safe_load(k8s_cluster.kubeconfig_path.read_text())
        user = kubeconfig["users"][0]["user"]
        ca = kubeconfig["clusters"][0]["cluster"].pop("certificate-authority-data")
        with tempfile.TemporaryDirectory() as d:
            kubeconfig["users"][0]["user"] = {
                "client-certificate": f"{d}/client.crt" if absolute else "client.crt",
                "client-key": f"{d}/client.key" if absolute else "client.key",
            }
            kubeconfig["clusters"][0]["cluster"]["certificate-authority"] = (
                f"{d}/ca.crt" if absolute else "ca.crt"
            )
            with open(f"{d}/client.crt", "wb") as f:
                f.write(base64.b64decode(user["client-certificate-data"]))
            with open(f"{d}/client.key", "wb") as f:
                f.write(base64.b64decode(user["client-key-data"]))
            with open(f"{d}/ca.crt", "wb") as f:
                f.write(base64.b64decode(ca))
            with open(f"{d}/config", "wb") as f:
                f.write(yaml.safe_dump(kubeconfig).encode())
                f.flush()
            yield f"{d}/config"

    return f


async def test_kubeconfig(k8s_cluster):
    api = await kr8s.asyncio.api(kubeconfig=k8s_cluster.kubeconfig_path)
    assert await anext(api.get("pods", namespace=kr8s.ALL))
    assert await api.whoami() == "kubernetes-admin"


async def test_kubeconfig_multi_paths_same(k8s_cluster):
    kubeconfig_multi_str = (
        f"{k8s_cluster.kubeconfig_path}:{k8s_cluster.kubeconfig_path}"
    )
    api = await kr8s.asyncio.api(kubeconfig=kubeconfig_multi_str)
    assert await anext(api.get("pods", namespace=kr8s.ALL))
    assert await api.whoami() == "kubernetes-admin"


async def test_kubeconfig_multi_paths_diff(k8s_cluster, tmp_path):
    kubeconfig1 = k8s_cluster.kubeconfig_path
    kubeconfig2 = Path(tmp_path / "kubeconfig").write_bytes(kubeconfig1.read_bytes())
    kubeconfig_multi_str = f"{kubeconfig1}:{kubeconfig2}"
    api = await kr8s.asyncio.api(kubeconfig=kubeconfig_multi_str)
    assert await anext(api.get("pods", namespace=kr8s.ALL))
    assert await api.whoami() == "kubernetes-admin"


async def test_kubeconfig_dict(k8s_cluster):
    config = yaml.safe_load(k8s_cluster.kubeconfig_path.read_text())
    assert isinstance(config, dict)
    api = await kr8s.asyncio.api(kubeconfig=config)
    assert await anext(api.get("pods", namespace=kr8s.ALL))
    assert await api.whoami() == "kubernetes-admin"


def test_kubeconfig_dict_sync(k8s_cluster):
    config = yaml.safe_load(k8s_cluster.kubeconfig_path.read_text())
    assert isinstance(config, dict)
    api = kr8s.api(kubeconfig=config)
    assert api.get("pods", namespace=kr8s.ALL)
    assert api.whoami() == "kubernetes-admin"


async def test_kubeconfig_context(kubeconfig_with_second_context):
    kubeconfig_path, context_name = kubeconfig_with_second_context
    api = await kr8s.asyncio.api(kubeconfig=kubeconfig_path, context=context_name)
    assert api.auth.active_context == context_name
    assert await anext(api.get("pods", namespace=kr8s.ALL))


async def test_default_service_account(k8s_cluster):
    api = await kr8s.asyncio.api(kubeconfig=k8s_cluster.kubeconfig_path)
    assert (
        str(api.auth._serviceaccount) == "/var/run/secrets/kubernetes.io/serviceaccount"
    )


async def test_reauthenticate(k8s_cluster):
    api = await kr8s.asyncio.api(kubeconfig=k8s_cluster.kubeconfig_path)
    await api.reauthenticate()
    assert await anext(api.get("pods", namespace=kr8s.ALL))


def test_reauthenticate_sync(k8s_cluster):
    api = kr8s.api(kubeconfig=k8s_cluster.kubeconfig_path)
    api.reauthenticate()
    assert api.get("pods", namespace=kr8s.ALL)


async def test_bad_auth(serviceaccount):
    (Path(serviceaccount) / "token").write_text("abc123")
    api = await kr8s.asyncio.api(
        serviceaccount=serviceaccount, kubeconfig="/no/file/here"
    )
    serviceaccount = Path(serviceaccount)
    with pytest.raises(kr8s.ServerError, match="Unauthorized"):
        await api.version()


async def test_url(kubectl_proxy):
    api = await kr8s.asyncio.api(url=kubectl_proxy)
    assert await anext(api.get("pods", namespace="kube-system"))
    assert api.auth.server == kubectl_proxy

    # Ensure reauthentication works
    api.auth.server = None
    await api.reauthenticate()
    assert await anext(api.get("pods", namespace="kube-system"))
    assert api.auth.server == kubectl_proxy


def test_no_config():
    with pytest.raises(ValueError):
        kr8s.api(kubeconfig="/no/file/here")


def test_kubeconfig_isdir_fail(tmp_path):
    with pytest.raises(IsADirectoryError):
        kr8s.api(kubeconfig=tmp_path)


async def test_service_account(serviceaccount):
    api = await kr8s.asyncio.api(
        serviceaccount=serviceaccount, kubeconfig="/no/file/here"
    )
    await api.version()

    serviceaccount = Path(serviceaccount)
    assert api.auth.server
    assert api.auth.token == (serviceaccount / "token").read_text()
    assert str(serviceaccount) in api.auth.server_ca_file
    assert "BEGIN CERTIFICATE" in Path(api.auth.server_ca_file).read_text()
    assert api.auth.namespace == (serviceaccount / "namespace").read_text()


async def test_service_account_with_kubeconfig_namespace(serviceaccount):
    kubeconfig = await KubeConfig(
        {
            "apiVersion": "v1",
            "clusters": None,
            "contexts": [
                {
                    "context": {"cluster": "", "namespace": "bar", "user": ""},
                    "name": "foo",
                }
            ],
            "current-context": "foo",
            "kind": "Config",
            "preferences": {},
            "users": None,
        }
    )
    kubeconfig_path = str(Path(serviceaccount) / "kubeconfig")
    await kubeconfig.save(path=kubeconfig_path)
    api = await kr8s.asyncio.api(
        serviceaccount=serviceaccount, kubeconfig=kubeconfig_path
    )

    assert api.auth.server
    assert api.namespace == "bar"


async def test_exec(kubeconfig_with_exec):
    api = await kr8s.asyncio.api(kubeconfig=kubeconfig_with_exec)
    assert await anext(api.get("pods", namespace=kr8s.ALL))
    assert api.auth.server
    assert api.auth.server_ca_file

    # Test reauthentication
    api.auth.server = None
    api.auth.server_ca_file = None
    await api.reauthenticate()
    assert api.auth.server
    assert api.auth.server_ca_file


async def test_token(kubeconfig_with_token):
    api = await kr8s.asyncio.api(kubeconfig=kubeconfig_with_token)
    assert await api.whoami() == "system:serviceaccount:default:pytest"
    assert await anext(api.get("pods", namespace=kr8s.ALL))


@pytest.mark.parametrize("absolute", [True, False])
async def test_certs_on_disk(kubeconfig_with_certs_on_disk, absolute):
    with kubeconfig_with_certs_on_disk(absolute=absolute) as kubeconfig:
        api = await kr8s.asyncio.api(kubeconfig=kubeconfig)
        assert await anext(api.get("pods", namespace=kr8s.ALL))


async def test_certs_not_encoded(kubeconfig_with_decoded_certs):
    api = await kr8s.asyncio.api(kubeconfig=kubeconfig_with_decoded_certs)
    assert await anext(api.get("pods", namespace=kr8s.ALL))


async def test_certs_with_encoded_line_breaks(kubeconfig_with_line_breaks_in_certs):
    api = await kr8s.asyncio.api(kubeconfig=kubeconfig_with_line_breaks_in_certs)
    assert await anext(api.get("pods", namespace=kr8s.ALL))


@pytest.mark.parametrize(
    "host,port,expected",
    [
        ("localhost", "8080", "https://localhost:8080"),
        ("9.9.9.9", "1234", "https://9.9.9.9:1234"),
        ("fd97:3495:4300::1", "443", "https://[fd97:3495:4300::1]:443"),
        ("kubernetes.default.svc", "8080", "https://kubernetes.default.svc:8080"),
    ],
)
def test_url_formatting(host, port, expected):
    assert KubeAuth._format_server_address(host, port) == expected



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
from __future__ import annotations

import asyncio
import contextlib
import copy
import json
import logging
import ssl
import threading
import warnings
import weakref
from typing import (
    TYPE_CHECKING,
    AsyncGenerator,
)

import httpx
import httpx_ws
from asyncache import cached  # type: ignore
from cachetools import TTLCache  # type: ignore
from cryptography import x509

from ._auth import KubeAuth
from ._data_utils import dict_to_selector, sort_versions
from ._exceptions import APITimeoutError, ServerError

if TYPE_CHECKING:
    from ._objects import APIObject

ALL = "all"
logger = logging.getLogger(__name__)


class Api:
    """A kr8s object for interacting with the Kubernetes API.

    .. warning::
        This class is not intended to be instantiated directly. Instead, use the
        :func:`kr8s.api` function to get a singleton instance of the API.

        See https://docs.kr8s.org/en/stable/client.html#client-caching.

    """

    _asyncio = True
    _instances: dict[str, weakref.WeakValueDictionary] = {}

    def __init__(self, **kwargs) -> None:
        if not kwargs.pop("bypass_factory", False):
            raise ValueError(
                "Use kr8s.api() to get an instance of Api. "
                "See https://docs.kr8s.org/en/stable/client.html#client-caching."
            )

        self._url = kwargs.get("url")
        self._kubeconfig = kwargs.get("kubeconfig")
        self._serviceaccount = kwargs.get("serviceaccount")
        self._session: httpx.AsyncClient | None = None
        self._timeout = None
        self.auth = KubeAuth(
            url=self._url,
            kubeconfig=self._kubeconfig,
            serviceaccount=self._serviceaccount,
            namespace=kwargs.get("namespace"),
            context=kwargs.get("context"),
        )
        thread_id = threading.get_ident()
        try:
            loop_id = id(asyncio.get_running_loop())
        except RuntimeError:
            loop_id = 0
        thread_loop_id = f"{thread_id}.{loop_id}"
        if thread_loop_id not in Api._instances:
            Api._instances[thread_loop_id] = weakref.WeakValueDictionary()
        key = hash_kwargs(kwargs)
        Api._instances[thread_loop_id][key] = self

    def __await__(self):
        async def f():
            await self.auth
            return self

        return f().__await__()

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        self._timeout = value
        if self._session:
            self._session.timeout = value

    async def _create_session(self) -> None:
        headers = {"User-Agent": self.__version__, "content-type": "application/json"}
        if self.auth.token:
            headers["Authorization"] = f"Bearer {self.auth.token}"
        if self._session:
            with contextlib.suppress(RuntimeError):
                await self._session.aclose()
            self._session = None
        self._session = httpx.AsyncClient(
            base_url=self.auth.server,
            headers=headers,
            verify=await self.auth.ssl_context(),
            timeout=self._timeout,
            follow_redirects=True,
        )

    def _construct_url(
        self,
        version: str = "v1",
        base: str = "",
        namespace: str | None = None,
        url: str = "",
    ) -> str:
        if not base:
            if version == "v1":
                base = "/api"
            elif "/" in version:
                base = "/apis"
            else:
                raise ValueError("Unknown API version, base must be specified.")
        parts = [base]
        if version:
            parts.append(version)
        if namespace:
            parts.extend(["namespaces", namespace])
        parts.append(url)
        return "/".join(parts)

    @contextlib.asynccontextmanager
    async def call_api(
        self,
        method: str = "GET",
        version: str = "v1",
        base: str = "",
        namespace: str | None = None,
        url: str = "",
        raise_for_status: bool = True,
        stream: bool = False,
        **kwargs,
    ) -> AsyncGenerator[httpx.Response]:
        """Make a Kubernetes API request."""
        if not self._session or self._session.is_closed:
            await self._create_session()
        url = self._construct_url(version, base, namespace, url)
        kwargs.update(url=url, method=method)
        if self.auth.tls_server_name:
            kwargs["extensions"] = {"sni_hostname": self.auth.tls_server_name}
        auth_attempts = 0
        ssl_attempts = 0
        while True:
            try:
                if stream:
                    assert self._session
                    async with self._session.stream(**kwargs) as response:
                        if raise_for_status:
                            response.raise_for_status()
                        yield response
                else:
                    assert self._session
                    response = await self._session.request(**kwargs)
                    if raise_for_status:
                        response.raise_for_status()
                    yield response
            except httpx.HTTPStatusError as e:
                # If we get a 401 or 403 our credentials may have expired so we
                # reauthenticate and try again a few times before giving up.
                if e.response.status_code in (401, 403) and auth_attempts < 3:
                    logger.debug(
                        f"Unauthorized {e.response.status_code} error, reauthenticating attempt {auth_attempts}"
                    )
                    auth_attempts += 1
                    await self.auth.reauthenticate()
                    await self._create_session()
                    continue
                else:
                    if e.response.status_code >= 400 and e.response.status_code < 500:
                        try:
                            error = e.response.json()
                            error_message = error["message"]
                        except json.JSONDecodeError:
                            error = e.response.text
                            error_message = str(e)
                        raise ServerError(
                            error_message, status=error, response=e.response
                        ) from e
                    elif e.response.status_code >= 500:
                        raise ServerError(
                            str(e),
                            status=str(e.response.status_code),
                            response=e.response,
                        ) from e
                    raise
            except ssl.SSLCertVerificationError:
                # In some rare edge cases the SSL verification fails, so we try again
                # a few times before giving up.
                if ssl_attempts < 3:
                    ssl_attempts += 1
                    await self.auth.reauthenticate()
                    await self._create_session()
                    continue
                else:
                    raise
            except httpx.TimeoutException as e:
                raise APITimeoutError(
                    "Timeout while waiting for the Kubernetes API server"
                ) from e
            break

    @contextlib.asynccontextmanager
    async def open_websocket(
        self,
        version: str = "v1",
        base: str = "",
        namespace: str | None = None,
        url: str = "",
        **kwargs,
    ) -> AsyncGenerator[httpx_ws.AsyncWebSocketSession]:
        """Open a websocket connection to a Kubernetes API endpoint."""
        if not self._session or self._session.is_closed:
            await self._create_session()
        url = self._construct_url(version, base, namespace, url)
        kwargs.update(url=url)
        if self.auth.tls_server_name:
            kwargs["extensions"] = {"sni_hostname": self.auth.tls_server_name}
        auth_attempts = 0
        while True:
            try:
                async with httpx_ws.aconnect_ws(
                    client=self._session, **kwargs
                ) as response:
                    yield response
            except httpx_ws.WebSocketDisconnect as e:
                if e.code and e.code != 1000:
                    if e.code in (401, 403) and auth_attempts < 3:
                        auth_attempts += 1
                        await self.auth.reauthenticate()
                        continue
                    else:
                        raise
            break

    async def version(self) -> dict:
        """Get the Kubernetes version information from the API.

        Returns:
            The Kubernetes version information.

        """
        return await self.async_version()

    async def async_version(self) -> dict:
        async with self.call_api(method="GET", version="", base="/version") as response:
            return response.json()

    async def reauthenticate(self) -> None:
        """Reauthenticate the API."""
        return await self.async_reauthenticate()

    async def async_reauthenticate(self) -> None:
        await self.auth.reauthenticate()

    async def whoami(self):
        """Retrieve the subject that's currently authenticated.

        Inspired by `kubectl whoami`.

        Returns:
            str: The subject that's currently authenticated.
        """
        return await self.async_whoami()

    async def async_whoami(self):
        if self.auth.token:
            payload = {
                "apiVersion": "authentication.k8s.io/v1",
                "kind": "TokenReview",
                "spec": {"token": self.auth.token},
            }
            async with self.call_api(
                "POST",
                version="authentication.k8s.io/v1",
                url="tokenreviews",
                data=json.dumps(payload),
            ) as r:
                data = r.json()
                return data["status"]["user"]["username"]
        elif self.auth.client_cert_file:
            with open(self.auth.client_cert_file, "rb") as f:
                cert = x509.load_pem_x509_certificate(f.read())
                [name] = cert.subject.get_attributes_for_oid(x509.OID_COMMON_NAME)
                return name.value

    async def lookup_kind(self, kind) -> tuple[str, str, bool]:
        """Lookup a Kubernetes resource kind.

        Check whether a resource kind exists on the remote server.

        Args:
            kind: The kind of resource to lookup.

        Returns:
            The kind of resource, the plural form and whether the resource is namespaced

        Raises:
            ValueError: If the kind is not found.
        """
        return await self.async_lookup_kind(kind)

    async def async_lookup_kind(self, kind) -> tuple[str, str, bool]:
        """Lookup a Kubernetes resource kind."""
        from ._objects import parse_kind

        resources = await self.async_api_resources()
        kind, group, version = parse_kind(kind)
        if group:
            version = f"{group}/{version}"
        for resource in resources:
            if (not version or version in resource["version"]) and (
                kind == resource["name"]
                or kind == resource["kind"]
                or kind == resource["singularName"]
                or ("shortNames" in resource and kind in resource["shortNames"])
            ):
                if "/" in resource["version"]:
                    return (
                        f"{resource['singularName']}.{resource['version']}",
                        resource["name"],
                        resource["namespaced"],
                    )
                return (
                    f"{resource['singularName']}/{resource['version']}",
                    resource["name"],
                    resource["namespaced"],
                )
        raise ValueError(f"Kind {kind} not found.")

    @contextlib.asynccontextmanager
    async def async_get_kind(
        self,
        kind: str | type[APIObject],
        namespace: str | None = None,
        label_selector: str | dict | None = None,
        field_selector: str | dict | None = None,
        params: dict | None = None,
        watch: bool = False,
        allow_unknown_type: bool = True,
        **kwargs,
    ) -> AsyncGenerator[tuple[type[APIObject], httpx.Response]]:
        """Get a Kubernetes resource."""
        from ._objects import get_class, new_class

        if not namespace:
            namespace = self.namespace
        if namespace is ALL:
            namespace = ""
        if params is None:
            params = {}
        if label_selector:
            if isinstance(label_selector, dict):
                label_selector = dict_to_selector(label_selector)
            params["labelSelector"] = label_selector
        if field_selector:
            if isinstance(field_selector, dict):
                field_selector = dict_to_selector(field_selector)
            params["fieldSelector"] = field_selector
        if watch:
            params["watch"] = "true" if watch else "false"
            kwargs["stream"] = True
        if isinstance(kind, type):
            obj_cls = kind
        else:
            namespaced: bool | None = None
            try:
                kind, plural, namespaced = await self.async_lookup_kind(kind)
            except ServerError as e:
                warnings.warn(str(e), stacklevel=1)
            if isinstance(kind, str):
                try:
                    obj_cls = get_class(kind, _asyncio=self._asyncio)
                except KeyError as e:
                    if allow_unknown_type:
                        if namespaced is not None:
                            obj_cls = new_class(
                                kind,
                                namespaced=namespaced,
                                asyncio=self._asyncio,
                                plural=plural,
                            )
                        else:
                            obj_cls = new_class(
                                kind, asyncio=self._asyncio, plural=plural
                            )
                    else:
                        raise e
        params = params or None
        async with self.call_api(
            method="GET",
            url=obj_cls.endpoint,
            version=obj_cls.version,
            namespace=namespace if obj_cls.namespaced else None,
            params=params,
            **kwargs,
        ) as response:
            yield obj_cls, response

    async def get(
        self,
        kind: str | type,
        *names: str,
        namespace: str | None = None,
        label_selector: str | dict | None = None,
        field_selector: str | dict | None = None,
        as_object: type[APIObject] | None = None,
        allow_unknown_type: bool = True,
        **kwargs,
    ) -> AsyncGenerator[APIObject]:
        """Get Kubernetes resources.

        Args:
            kind: The kind of resource to get.
            *names: The names of specific resources to get.
            namespace: The namespace to get the resource from.
            label_selector: The label selector to filter the resources by.
            field_selector: The field selector to filter the resources by.
            as_object: The object to return the resources as.
            allow_unknown_type: Automatically create a class for the resource if none exists, default True.
            **kwargs: Additional keyword arguments to pass to the API call.

        Returns:
            The resources.
        """
        async for resource in self.async_get(
            kind,
            *names,
            namespace=namespace,
            label_selector=label_selector,
            field_selector=field_selector,
            as_object=as_object,
            allow_unknown_type=allow_unknown_type,
            **kwargs,
        ):
            yield resource

    async def async_get(
        self,
        kind: str | type,
        *names: str,
        namespace: str | None = None,
        label_selector: str | dict | None = None,
        field_selector: str | dict | None = None,
        as_object: type[APIObject] | None = None,
        allow_unknown_type: bool = True,
        **kwargs,
    ) -> AsyncGenerator[APIObject]:
        names_list = [None] if not names else names
        for name in names_list:
            async for resource in self._async_get_single(
                kind,
                name,
                namespace=namespace,
                label_selector=label_selector,
                field_selector=field_selector,
                as_object=as_object,
                allow_unknown_type=allow_unknown_type,
                **kwargs,
            ):
                yield resource

    async def _async_get_single(
        self,
        kind: str | type,
        name: str | None = None,
        namespace: str | None = None,
        label_selector: str | dict | None = None,
        field_selector: str | dict | None = None,
        as_object: type[APIObject] | None = None,
        allow_unknown_type: bool = True,
        **kwargs,
    ) -> AsyncGenerator[APIObject]:

        if name is not None:
            # Normalized field_selector to a string
            field_selector_str: str
            if isinstance(field_selector, dict):
                field_selector_str = dict_to_selector(field_selector)
            elif field_selector is None:
                field_selector_str = ""
            else:
                field_selector_str = field_selector
            field_selector = f"metadata.name={name},{field_selector_str}"

        headers = {}
        params = {}
        continue_paging = True
        if as_object:
            group, version = as_object.version.split("/")
            headers["Accept"] = (
                f"application/json;as={as_object.kind};v={version};g={group}"
            )
        else:
            params["limit"] = 100
        while continue_paging:
            async with self.async_get_kind(
                kind,
                namespace=namespace,
                label_selector=label_selector,
                field_selector=field_selector,
                headers=headers or None,
                allow_unknown_type=allow_unknown_type,
                params=params,
                **kwargs,
            ) as (obj_cls, response):
                resourcelist = response.json()
                if (
                    as_object
                    and "kind" in resourcelist
                    and resourcelist["kind"] == as_object.kind
                ):
                    yield as_object(resourcelist, api=self)
                else:
                    if "items" in resourcelist:
                        for item in resourcelist["items"]:
                            if name is None or item["metadata"]["name"] == name:
                                yield obj_cls(item, api=self)
                if (
                    "metadata" in resourcelist
                    and "continue" in resourcelist["metadata"]
                    and resourcelist["metadata"]["continue"]
                ):
                    continue_paging = True
                    params["continue"] = resourcelist["metadata"]["continue"]
                else:
                    continue_paging = False

    async def watch(
        self,
        kind: str,
        namespace: str | None = None,
        label_selector: str | dict | None = None,
        field_selector: str | dict | None = None,
        since: str | None = None,
    ) -> AsyncGenerator[tuple[str, APIObject]]:
        """Watch a Kubernetes resource."""
        async for t, object in self.async_watch(
            kind,
            namespace=namespace,
            label_selector=label_selector,
            field_selector=field_selector,
            since=since,
        ):
            yield t, object

    async def async_watch(
        self,
        kind: str,
        namespace: str | None = None,
        label_selector: str | dict | None = None,
        field_selector: str | dict | None = None,
        since: str | None = None,
        allow_unknown_type: bool = True,
    ) -> AsyncGenerator[tuple[str, APIObject]]:
        """Watch a Kubernetes resource."""
        while True:
            restart_watch = False
            async with self.async_get_kind(
                kind,
                namespace=namespace,
                label_selector=label_selector,
                field_selector=field_selector,
                params={"resourceVersion": since} if since else None,
                watch=True,
                timeout=None,
                allow_unknown_type=allow_unknown_type,
            ) as (obj_cls, response):
                logger.debug(
                    f"Starting watch of {kind}{' at resourceVersion ' + since if since else ''}"
                )
                async for line in response.aiter_lines():
                    event = json.loads(line)
                    if (
                        event["object"]["kind"] == "Status"
                        and event["object"].get("code") == 410
                    ):
                        restart_watch = True
                        logger.debug(
                            f"Got 410 Gone: Restarting watch of {kind} at resourceVersion {since}"
                        )
                        break
                    obj = obj_cls(event["object"], api=self)
                    since = obj.metadata.resourceVersion
                    yield event["type"], obj
            if not restart_watch:
                return

    async def api_resources(self) -> list[dict]:
        """Get the Kubernetes API resources."""
        return await self.async_api_resources()

    # Cache for 6 hours because kubectl does
    # https://github.com/kubernetes/cli-runtime/blob/980bedf450ab21617b33d68331786942227fe93a/pkg/genericclioptions/config_flags.go#L297
    @cached(TTLCache(1, 60 * 60 * 6))
    async def async_api_resources(self) -> list[dict]:
        """Get the Kubernetes API resources."""
        resources = []
        async with self.call_api(method="GET", version="", base="/api") as response:
            core_api_list = response.json()

        for version in core_api_list["versions"]:
            async with self.call_api(
                method="GET", version="", base="/api", url=version
            ) as response:
                resource = response.json()
            resources.extend(
                [
                    {"version": version, **r}
                    for r in resource["resources"]
                    if "/" not in r["name"]
                ]
            )
        async with self.call_api(method="GET", version="", base="/apis") as response:
            api_list = response.json()
        for api in sorted(api_list["groups"], key=lambda d: d["name"]):
            for api_version in sort_versions(
                api["versions"], key=lambda x: x["groupVersion"]
            ):
                version = api_version["groupVersion"]
                async with self.call_api(
                    method="GET", version="", base="/apis", url=version
                ) as response:
                    resource = response.json()
                resources.extend(
                    [
                        {"version": version, **r}
                        for r in resource["resources"]
                        if "/" not in r["name"]
                    ]
                )
        return resources

    async def api_versions(self) -> AsyncGenerator[str]:
        """Get the Kubernetes API versions."""
        async for version in self.async_api_versions():
            yield version

    async def async_api_versions(self) -> AsyncGenerator[str]:
        async with self.call_api(method="GET", version="", base="/api") as response:
            core_api_list = response.json()
        for version in core_api_list["versions"]:
            yield version

        async with self.call_api(method="GET", version="", base="/apis") as response:
            api_list = response.json()
        for group in api_list["groups"]:
            for version in group["versions"]:
                yield version["groupVersion"]

    @property
    def __version__(self) -> str:
        from . import __version__

        return f"kr8s/{__version__}"

    @property
    def namespace(self) -> str:
        """Get the default namespace."""
        assert self.auth.namespace
        return self.auth.namespace

    @namespace.setter
    def namespace(self, value):
        self.auth.namespace = value


def hash_kwargs(kwargs: dict):
    key_kwargs = copy.copy(kwargs)
    for key in key_kwargs:
        if isinstance(key_kwargs[key], dict):
            key_kwargs[key] = json.dumps(key_kwargs[key])
    return frozenset(key_kwargs.items())



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
"""The `kr8s` asynchronous API.

This module provides an asynchronous API for interacting with a Kubernetes cluster.
"""
from kr8s._api import Api

from . import objects, portforward
from ._api import api
from ._helpers import api_resources, get, version, watch, whoami

__all__ = [
    "api",
    "api_resources",
    "get",
    "objects",
    "portforward",
    "version",
    "watch",
    "whoami",
    "Api",
]



# SPDX-FileCopyrightText: Copyright (c) 2024-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
"""Objects for managing a port forward connection.

This module provides a class for managing a port forward connection to a Kubernetes Pod or Service.
"""
# Disable missing docstrings, these are inherited from the async version of the objects
# ruff: noqa: D102, D105
from __future__ import annotations

import threading
import time

from ._async_utils import run_sync
from ._portforward import LocalPortType
from ._portforward import PortForward as _PortForward

__all__ = ["PortForward", "LocalPortType"]


class PortForward(_PortForward):
    _bg_thread = None

    def __enter__(self, *args, **kwargs):
        return run_sync(self.__aenter__)(*args, **kwargs)

    def __exit__(self, *args, **kwargs):
        return run_sync(self.__aexit__)(*args, **kwargs)

    def run_forever(self):
        return run_sync(self.run_forever)()  # type: ignore

    def start(self):
        """Start a background thread with the port forward running."""
        self._bg_thread = threading.Thread(target=self.run_forever, daemon=True)
        self._bg_thread.start()

    def stop(self):
        """Stop the background thread."""
        for server in self.servers:
            while server is None:
                time.sleep(0.1)
            server.close()



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
from kubectl_ng.cli import app
from typer.testing import CliRunner

runner = CliRunner()


def test_version():
    result = runner.invoke(app, ["api-versions"])
    assert result.exit_code == 0
    assert "apps/v1" in result.stdout



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
from __future__ import annotations

import contextlib
import json
import pathlib
import re
import time
from typing import (
    Any,
    AsyncGenerator,
    BinaryIO,
    List,
    Literal,
    cast,
)

import anyio
import httpx
import jsonpath
import yaml
from box import Box
from typing_extensions import Self

import kr8s
import kr8s.asyncio
from kr8s._api import Api
from kr8s._async_utils import run_sync
from kr8s._data_utils import (
    dict_to_selector,
    dot_to_nested_dict,
    list_dict_unpack,
    xdict,
)
from kr8s._exceptions import NotFoundError, ServerError
from kr8s._exec import Exec
from kr8s._types import SpecType, SupportsKeysAndGetItem
from kr8s.asyncio.portforward import PortForward as AsyncPortForward
from kr8s.portforward import LocalPortType
from kr8s.portforward import PortForward as SyncPortForward

JSONPATH_CONDITION_EXPRESSION = r"jsonpath='{(?P<expression>.*?)}'=(?P<condition>.*)"


class APIObject:
    """Base class for Kubernetes objects."""

    version: str
    endpoint: str
    kind: str
    plural: str
    singular: str
    namespaced: bool = False
    scalable: bool = False
    scalable_spec: str = "replicas"
    _asyncio: bool = True

    def __init__(
        self, resource: SpecType, namespace: str | None = None, api: Api | None = None
    ) -> None:
        """Initialize an APIObject."""
        if isinstance(resource, dict):
            self.raw = resource
        elif isinstance(resource, SupportsKeysAndGetItem):
            self.raw = dict(resource)
        elif isinstance(resource, str):
            self.raw = {"metadata": {"name": resource}}
        elif hasattr(resource, "to_dict"):
            self.raw = resource.to_dict()
        elif hasattr(resource, "obj") and isinstance(resource.obj, dict):
            self.raw = resource.obj
        else:
            raise ValueError(
                "resource must be a dict, string, have a to_dict method or an obj attribute containing a dict"
            )
        if namespace is not None:
            self.raw["metadata"]["namespace"] = namespace  # type: ignore
        self._api = api
        if self._api is None and not self._asyncio:
            self._api = kr8s.api()

    def __await__(self):
        async def f():
            if self._api is None:
                self._api = await kr8s.asyncio.api()
            return self

        return f().__await__()

    def __repr__(self):
        """Return a string representation of the Kubernetes resource."""
        if "name" in self.metadata:
            return f"<{self.kind} {self.name}>"
        if "generateName" in self.metadata:
            return f"<{self.kind} generateName({self.metadata.generateName})>"
        return f"<{self.kind} UNKNOWN>"

    def __str__(self):
        """Return a string representation of the Kubernetes resource."""
        return self.name

    def __eq__(self, other):
        if self.version != other.version:
            return False
        if self.kind != other.kind:
            return False
        if self.name != other.name:
            return False
        if self.namespaced and self.namespace != other.namespace:
            return False
        return True

    @property
    def api(self):
        if self._api is None and self._asyncio:
            raise RuntimeError(
                f"{self.__repr__()} has not been initialized with an API object. "
                "Did you forget to await it?"
            )
        return self._api

    @api.setter
    def api(self, value):
        self._api = value

    @property
    def raw(self) -> Any:
        """Raw object returned from the Kubernetes API."""
        self._raw.update({"kind": self.kind, "apiVersion": self.version})
        return self._raw

    @raw.setter
    def raw(self, value: Any) -> None:
        self._raw = Box(value)

    @property
    def name(self) -> str:
        """Name of the Kubernetes resource."""
        if "name" in self.metadata:
            return self.metadata.name
        if "generateName" in self.metadata:
            raise ValueError("Resource has a generateName that has not been resolved")
        raise ValueError("Resource does not have a name")

    @name.setter
    def name(self, value: str) -> None:
        self.raw["metadata"]["name"] = value

    @property
    def namespace(self) -> str | None:
        """Namespace of the Kubernetes resource."""
        if self.namespaced:
            assert self.api
            return self.raw.get("metadata", {}).get("namespace", self.api.namespace)
        return None

    @namespace.setter
    def namespace(self, value: str) -> None:
        if self.namespaced:
            self.raw["metadata"]["namespace"] = value

    @property
    def metadata(self) -> Box:
        """Metadata of the Kubernetes resource."""
        return self.raw["metadata"]

    @metadata.setter
    def metadata(self, value: dict) -> None:
        self.raw["metadata"] = value

    @property
    def spec(self) -> Box:
        """Spec of the Kubernetes resource."""
        return self.raw["spec"]

    @spec.setter
    def spec(self, value: dict) -> None:
        self.raw["spec"] = value

    @property
    def status(self) -> Box:
        """Status of the Kubernetes resource."""
        return self.raw["status"]

    @status.setter
    def status(self, value: dict) -> None:
        self.raw["status"] = value

    @property
    def labels(self) -> Box:
        """Labels of the Kubernetes resource."""
        try:
            return self.raw["metadata"]["labels"]
        except KeyError:
            return Box({})

    @labels.setter
    def labels(self, value: dict) -> None:
        self.raw["metadata"]["labels"] = value

    @property
    def annotations(self) -> Box:
        """Annotations of the Kubernetes resource."""
        try:
            return self.raw["metadata"]["annotations"]
        except KeyError:
            return Box({})

    @annotations.setter
    def annotations(self, value: dict) -> None:
        self.raw["metadata"]["annotations"] = value

    @property
    def replicas(self) -> int:
        """Replicas of the Kubernetes resource."""
        if self.scalable:
            keys = self.scalable_spec.split(".")
            spec = self.raw["spec"]
            for key in keys:
                spec = spec[key]
            return spec
        raise NotImplementedError(f"{self.kind} is not scalable")

    @replicas.setter
    def replicas(self, value: int) -> None:
        if self.scalable:
            raise NotImplementedError(
                f"Use scale() method to set replicas for {self.kind}"
            )
        else:
            raise NotImplementedError(f"{self.kind} is not scalable")

    @classmethod
    async def get(
        cls,
        name: str | None = None,
        namespace: str | None = None,
        api: Api | None = None,
        label_selector: str | dict[str, str] | None = None,
        field_selector: str | dict[str, str] | None = None,
        timeout: int = 2,
        **kwargs,
    ) -> Self:
        """Get a Kubernetes resource by name or via selectors."""
        return await cls.async_get(
            name=name,
            namespace=namespace,
            api=api,
            label_selector=label_selector,
            field_selector=field_selector,
            timeout=timeout,
            **kwargs,
        )

    @classmethod
    async def async_get(
        cls,
        name: str | None = None,
        namespace: str | None = None,
        api: Api | None = None,
        label_selector: str | dict[str, str] | None = None,
        field_selector: str | dict[str, str] | None = None,
        timeout: int = 2,
        **kwargs,
    ) -> Self:
        """Get a Kubernetes resource by name or via selectors."""
        if api is None:
            if cls._asyncio:
                api = await kr8s.asyncio.api()
            else:
                api = await kr8s.asyncio.api(_asyncio=False)
        namespace = namespace if namespace else api.namespace
        start = time.time()
        backoff = 0.1
        while start + timeout > time.time():
            if name:
                try:
                    resources = [
                        resource
                        async for resource in api.async_get(
                            cls,
                            name,
                            namespace=namespace,
                            field_selector={"metadata.name": name},
                            **kwargs,
                        )
                    ]
                except ServerError as e:
                    if e.response and e.response.status_code == 404:
                        continue
                    raise e
            elif label_selector or field_selector:
                resources = [
                    resource
                    async for resource in api.async_get(
                        cls,
                        namespace=namespace,
                        label_selector=label_selector,
                        field_selector=field_selector,
                        **kwargs,
                    )
                ]
            else:
                raise ValueError("Must specify name or selector")
            if not isinstance(resources, list):
                resources = [resources]
            if len(resources) == 0:
                await anyio.sleep(backoff)
                backoff = min(backoff * 2, 1)
                continue
            if len(resources) > 1:
                raise ValueError(
                    f"Expected exactly one {cls.kind} object. Use selectors to narrow down the search."
                )
            assert isinstance(resources[0], cls)
            return resources[0]
        raise NotFoundError(
            f"Could not find {cls.kind} {name} in namespace {namespace}."
        )

    async def exists(self, ensure=False) -> bool:
        """Check if this object exists in Kubernetes."""
        return await self.async_exists(ensure=ensure)

    async def async_exists(self, ensure=False) -> bool:
        """Check if this object exists in Kubernetes."""
        try:
            assert self.api
            async with self.api.call_api(
                "GET",
                version=self.version,
                url=f"{self.endpoint}/{self.name}",
                namespace=self.namespace,
                raise_for_status=False,
            ) as resp:
                status = resp.status_code
        except ValueError:
            status = 400
        if status == 200:
            return True
        if ensure:
            raise NotFoundError(
                f"Object {self.name} does not exist in {self.namespace}"
            )
        return False

    async def async_create(self) -> None:
        """Create this object in Kubernetes."""
        assert self.api
        async with self.api.call_api(
            "POST",
            version=self.version,
            url=self.endpoint,
            namespace=self.namespace,
            data=json.dumps(self.raw),
        ) as resp:
            self.raw = resp.json()

    async def create(self) -> None:
        """Create this object in Kubernetes."""
        return await self.async_create()

    async def delete(self, propagation_policy: str | None = None) -> None:
        """Delete this object from Kubernetes."""
        return await self.async_delete(propagation_policy=propagation_policy)

    async def async_delete(self, propagation_policy: str | None = None) -> None:
        """Delete this object from Kubernetes."""
        data = {}
        if propagation_policy:
            data["propagationPolicy"] = propagation_policy
        try:
            assert self.api
            async with self.api.call_api(
                "DELETE",
                version=self.version,
                url=f"{self.endpoint}/{self.name}",
                namespace=self.namespace,
                data=json.dumps(data),
            ):
                pass
        except ServerError as e:
            if e.response and e.response.status_code == 404:
                raise NotFoundError(
                    f"Object {self.name} does not exist in {self.namespace}"
                ) from e
            raise e

    async def refresh(self) -> None:
        """Refresh this object from Kubernetes."""
        await self.async_refresh()

    async def async_refresh(self) -> None:
        """Refresh this object from Kubernetes."""
        try:
            assert self.api
            async with self.api.call_api(
                "GET",
                version=self.version,
                url=f"{self.endpoint}/{self.name}",
                namespace=self.namespace,
            ) as resp:
                self.raw = resp.json()
        except ServerError as e:
            if e.response and e.response.status_code == 404:
                raise NotFoundError(
                    f"Object {self.name} does not exist in {self.namespace}"
                ) from e
            raise e

    async def patch(self, patch, *, subresource=None, type=None) -> None:
        """Patch this object in Kubernetes."""
        await self.async_patch(patch, subresource=subresource, type=type)

    async def async_patch(
        self, patch: dict | list, *, subresource=None, type=None
    ) -> None:
        """Patch this object in Kubernetes."""
        url = f"{self.endpoint}/{self.name}"
        if type == "json":
            headers = {"Content-Type": "application/json-patch+json"}
        else:
            headers = {"Content-Type": "application/merge-patch+json"}
        if subresource:
            url = f"{url}/{subresource}"
        try:
            assert self.api
            async with self.api.call_api(
                "PATCH",
                version=self.version,
                url=url,
                namespace=self.namespace,
                data=json.dumps(patch),
                headers=headers,
            ) as resp:
                self.raw = resp.json()
        except ServerError as e:
            if e.response and e.response.status_code == 404:
                raise NotFoundError(
                    f"Object {self.name} does not exist in {self.namespace}"
                ) from e
            raise e

    async def scale(self, replicas: int | None = None) -> None:
        """Scale this object in Kubernetes."""
        return await self.async_scale(replicas=replicas)

    async def async_scale(self, replicas: int | None = None) -> None:
        """Scale this object in Kubernetes."""
        if not self.scalable:
            raise NotImplementedError(f"{self.kind} is not scalable")
        await self.async_exists(ensure=True)
        await self.async_patch(
            {"spec": dot_to_nested_dict(self.scalable_spec, replicas)}
        )
        while self.replicas != replicas:
            await self.async_refresh()
            await anyio.sleep(0.1)

    async def async_watch(self):
        """Watch this object in Kubernetes."""
        since = self.metadata.get("resourceVersion")
        assert self.api
        async for event, obj in self.api.async_watch(
            self.endpoint,
            namespace=self.namespace,
            field_selector=f"metadata.name={self.name}",
            since=since,
        ):
            self.raw = obj.raw
            yield event, self

    async def watch(self):
        """Watch this object in Kubernetes."""
        async for event, obj in self.async_watch():
            yield event, obj

    async def _test_conditions(
        self, conditions: list, mode: Literal["any", "all"] = "any"
    ) -> bool:
        """Test if conditions are met.

        Args:
            conditions: A list of conditions to test.
            mode: Match any condition with "any" or all conditions with "all". Defaults to "any".

        Returns:
            bool: True if any condition is met, False otherwise.
        """
        results = []
        for condition in conditions:
            if condition.startswith("condition"):
                condition = "=".join(condition.split("=")[1:])
                if "=" in condition:
                    field, value = condition.split("=")
                    value = str(value)
                else:
                    field = condition
                    value = str(True)
                if value == "true" or value == "false":
                    value = value.title()
                status_conditions = list_dict_unpack(
                    self.status.get("conditions", []), "type", "status"
                )
                results.append(status_conditions.get(field, None) == value)
            elif condition == "delete":
                results.append(not await self.async_exists())
            elif condition.startswith("jsonpath"):
                matches = re.search(JSONPATH_CONDITION_EXPRESSION, condition)
                if not matches:
                    raise ValueError(
                        f"Invalid jsonpath condition expression '{condition}'"
                    )
                expression = matches.group("expression")
                condition = matches.group("condition")
                [value] = jsonpath.findall(expression, self.raw)
                results.append(str(value) == str(condition))
            else:
                raise ValueError(f"Unknown condition type {condition}")
        if mode == "any":
            return any(results)
        elif mode == "all":
            return all(results)
        else:
            raise ValueError(f"Unknown wait mode '{mode}', must be 'any' or 'all'")

    async def wait(
        self,
        conditions: list[str] | str,
        mode: Literal["any", "all"] = "any",
        timeout: int | None = None,
    ):
        """Wait for conditions to be met.

        Args:
            conditions: A list of conditions to wait for.
            mode: Match any condition with "any" or all conditions with "all". Defaults to "any".
            timeout: Timeout in seconds.

        Example:
            Wait for a Pod to be ready.

            >>> from kr8s.objects import Pod
            >>> pod = Pod.get("my-pod")
            >>> pod.wait("condition=Ready")

            Wait for a Job to either succeed or fail.

            >>> from kr8s.objects import Job
            >>> job = Job.get("my-jod")
            >>> job.wait(["condition=Complete", "condition=Failed"])

            Wait for a Pod to be initialized and ready to start containers.

            >>> from kr8s.objects import Pod
            >>> pod = Pod.get("my-pod")
            >>> pod.wait(["condition=Initialized", "condition=PodReadyToStartContainers"], mode="all")

        Note:
            As of the current Kubertenetes release when this function was written (1.29) kubectl doesn't support
            multiple conditions. There is a PR to implement this but it hasn't been merged yet
            https://github.com/kubernetes/kubernetes/pull/119205.

            Given that ``for`` is a reserved word anyway we can't exactly match the kubectl API so
            we use ``condition`` in combination with a ``mode`` instead.
        """
        return await self.async_wait(conditions, mode=mode, timeout=timeout)

    async def async_wait(
        self,
        conditions: list[str] | str,
        mode: Literal["any", "all"] = "any",
        timeout: int | None = None,
    ):
        if isinstance(conditions, str):
            conditions = [conditions]

        with anyio.fail_after(timeout):
            try:
                await self.async_refresh()
            except NotFoundError:
                if set(conditions) == {"delete"}:
                    return
            if await self._test_conditions(conditions, mode=mode):
                return
            async for _ in self.async_watch():
                if await self._test_conditions(conditions, mode=mode):
                    return

    async def annotate(self, annotations: dict | None = None, **kwargs) -> None:
        """Annotate this object in Kubernetes."""
        return await self.async_annotate(annotations=annotations, **kwargs)

    async def async_annotate(self, annotations: dict | None = None, **kwargs) -> None:
        """Annotate this object in Kubernetes."""
        if annotations is None:
            annotations = kwargs
        if not annotations:
            raise ValueError("No annotations provided")
        await self.async_patch({"metadata": {"annotations": annotations}})

    async def label(self, labels: dict | None = None, **kwargs) -> None:
        """Add labels to this object in Kubernetes.

        Labels can be passed as a dictionary or as keyword arguments.

        Args:
            labels:
                A dictionary of labels to set.
            **kwargs:
                Labels to set.

        Example:
            >>> from kr8s.objects import Deployment
            >>> deployment = Deployment.get("my-deployment")
            >>> # Both of these are equivalent
            >>> deployment.label({"app": "my-app"})
            >>> deployment.label(app="my-app")
        """
        return await self.async_label(labels=labels, **kwargs)

    async def async_label(self, labels: dict | None = None, **kwargs) -> None:
        if labels is None:
            labels = kwargs
        if not labels:
            raise ValueError("No labels provided")
        await self.async_patch({"metadata": {"labels": labels}})

    def keys(self) -> list:
        """Return the keys of this object."""
        return self.raw.keys()

    def __getitem__(self, key: str) -> Any:
        """Get an item from this object."""
        return self.raw[key]

    def __setitem__(self, key: str, value: Any) -> None:
        """Set an item in this object."""
        self.raw[key] = value

    async def set_owner(self, owner: APIObject) -> None:
        """Set the owner reference of this object.

        See https://kubernetes.io/docs/concepts/overview/working-with-objects/owners-dependents/

        Args:
            owner: The owner object to set a reference to.

        Example:
            >>> from kr8s.objects import Deployment, Pod
            >>> deployment = Deployment.get("my-deployment")
            >>> pod = Pod.get("my-pod")
            >>> pod.set_owner(deployment)
        """
        await self.async_set_owner(owner)

    async def async_set_owner(self, owner: APIObject) -> None:
        """Set the owner of this object."""
        await self.async_patch(
            {
                "metadata": {
                    "ownerReferences": [
                        {
                            "controller": True,
                            "blockOwnerDeletion": True,
                            "apiVersion": owner.version,
                            "kind": owner.kind,
                            "name": owner.name,
                            "uid": owner.metadata.uid,
                        }
                    ],
                }
            }
        )

    async def adopt(self, child: APIObject) -> None:
        """Adopt this object.

        This will set the owner reference of the child object to this object.

        See https://kubernetes.io/docs/concepts/overview/working-with-objects/owners-dependents/

        Args:
            child: The child object to adopt.

        Example:
            >>> from kr8s.objects import Deployment, Pod
            >>> deployment = Deployment.get("my-deployment")
            >>> pod = Pod.get("my-pod")
            >>> deployment.adopt(pod)

        """
        return await self.async_adopt(child)

    async def async_adopt(self, child: APIObject) -> None:
        await child.async_set_owner(self)

    def to_dict(self) -> dict:
        """Return a dictionary representation of this object."""
        return self.raw

    def to_lightkube(self) -> Any:
        """Return a lightkube representation of this object."""
        try:
            from lightkube import codecs
        except ImportError as e:
            raise ImportError("lightkube is not installed") from e
        return codecs.from_dict(self.raw)

    def to_pykube(self, api) -> Any:
        """Return a pykube representation of this object.

        Args:
            api: A pykube API object.

        Example:
            >>> from kr8s.objects import Deployment
            >>> deployment = Deployment.get("my-deployment")
            >>> # Create a pykube API object
            >>> from pykube import HTTPClient
            >>> api = HTTPClient()
            >>> pykube_deployment = deployment.to_pykube(api)

        """
        try:
            import pykube  # type: ignore
        except ImportError as e:
            raise ImportError("pykube is not installed") from e
        try:
            pykube_cls = getattr(pykube.objects, self.kind)
        except AttributeError:
            base = (
                pykube.objects.NamespacedAPIObject
                if self.namespaced
                else pykube.objects.APIObject
            )
            pykube_cls = type(
                self.kind,
                (base,),
                {"version": self.version, "endpoint": self.endpoint, "kind": self.kind},
            )
        return pykube_cls(api, self.raw)

    @classmethod
    def gen(cls, *args, **kwargs):
        raise NotImplementedError("gen is not implemented for this object")

    @classmethod
    async def async_list(cls, **kwargs) -> AsyncGenerator[Self]:
        api = await kr8s.asyncio.api()
        async for resource in api.async_get(kind=cls, **kwargs):
            if isinstance(resource, cls):
                yield resource

    # Must be the last method defined due to https://github.com/python/mypy/issues/17517
    @classmethod
    async def list(cls, **kwargs) -> AsyncGenerator[Self]:
        """List objects in Kubernetes.

        Args:
            **kwargs: Keyword arguments to pass to :func:`kr8s.get`.

        Returns:
            A list of objects.
        """
        async for resource in cls.async_list(**kwargs):
            yield resource


class APIObjectSyncMixin(APIObject):
    _asyncio = False

    @classmethod
    def get(  # type: ignore
        cls,
        name: str | None = None,
        namespace: str | None = None,
        api: Api | None = None,
        label_selector: str | dict[str, str] | None = None,
        field_selector: str | dict[str, str] | None = None,
        timeout: int = 2,
        **kwargs,
    ) -> Self:
        return run_sync(cls.async_get)(
            name=name,
            namespace=namespace,
            api=api,
            label_selector=label_selector,
            field_selector=field_selector,
            timeout=timeout,
            **kwargs,
        )  # type: ignore

    def exists(self, ensure=False) -> bool:  # type: ignore
        return run_sync(self.async_exists)(ensure=ensure)  # type: ignore

    def create(self) -> None:  # type: ignore
        return run_sync(self.async_create)()  # type: ignore

    def delete(self, propagation_policy: str | None = None) -> None:  # type: ignore
        return run_sync(self.async_delete)(propagation_policy=propagation_policy)  # type: ignore

    def refresh(self):
        return run_sync(self.async_refresh)()  # type: ignore

    def patch(self, patch, *, subresource=None, type=None):
        return run_sync(self.async_patch)(patch, subresource=subresource, type=type)  # type: ignore

    def scale(self, replicas=None):
        return run_sync(self.async_scale)(replicas=replicas)  # type: ignore

    def watch(self):
        yield from run_sync(self.async_watch)()

    def wait(
        self,
        conditions: list[str] | str,
        mode: Literal["any", "all"] = "any",
        timeout: int | float | None = None,
    ):
        return run_sync(self.async_wait)(conditions, mode=mode, timeout=timeout)  # type: ignore

    def annotate(self, annotations=None, **kwargs):
        return run_sync(self.async_annotate)(annotations, **kwargs)  # type: ignore

    def label(self, labels=None, **kwargs):
        return run_sync(self.async_label)(labels, **kwargs)  # type: ignore

    def set_owner(self, owner):
        return run_sync(self.async_set_owner)(owner)  # type: ignore

    def adopt(self, child):
        return run_sync(self.async_adopt)(child)  # type: ignore

    @classmethod
    def list(cls):
        yield from run_sync(cls.async_list)()


## v1 objects


class Binding(APIObject):
    """A Kubernetes Binding."""

    version = "v1"
    endpoint = "bindings"
    kind = "Binding"
    plural = "bindings"
    singular = "binding"
    namespaced = True


class ComponentStatus(APIObject):
    """A Kubernetes ComponentStatus."""

    version = "v1"
    endpoint = "componentstatuses"
    kind = "ComponentStatus"
    plural = "componentstatuses"
    singular = "componentstatus"
    namespaced = False


class ConfigMap(APIObject):
    """A Kubernetes ConfigMap."""

    version = "v1"
    endpoint = "configmaps"
    kind = "ConfigMap"
    plural = "configmaps"
    singular = "configmap"
    namespaced = True

    @property
    def data(self) -> Box:
        """Data of the ConfigMap."""
        return self.raw["data"]

    @data.setter
    def data(self, value: dict) -> None:
        self.raw["data"] = value


class Endpoints(APIObject):
    """A Kubernetes Endpoints."""

    version = "v1"
    endpoint = "endpoints"
    kind = "Endpoints"
    plural = "endpoints"
    singular = "endpoint"
    namespaced = True


class Event(APIObject):
    """A Kubernetes Event."""

    version = "v1"
    endpoint = "events"
    kind = "Event"
    plural = "events"
    singular = "event"
    namespaced = True


class LimitRange(APIObject):
    """A Kubernetes LimitRange."""

    version = "v1"
    endpoint = "limitranges"
    kind = "LimitRange"
    plural = "limitranges"
    singular = "limitrange"
    namespaced = True


class Namespace(APIObject):
    """A Kubernetes Namespace."""

    version = "v1"
    endpoint = "namespaces"
    kind = "Namespace"
    plural = "namespaces"
    singular = "namespace"
    namespaced = False


class Node(APIObject):
    """A Kubernetes Node."""

    version = "v1"
    endpoint = "nodes"
    kind = "Node"
    plural = "nodes"
    singular = "node"
    namespaced = False

    @property
    def unschedulable(self):
        if "unschedulable" in self.raw["spec"]:
            return self.raw["spec"]["unschedulable"]
        return False

    async def cordon(self) -> None:
        """Cordon the node.

        This will mark the node as unschedulable.
        """
        return await self.async_cordon()

    async def async_cordon(self) -> None:
        await self.async_patch({"spec": {"unschedulable": True}})

    async def uncordon(self) -> None:
        """Uncordon the node.

        This will mark the node as schedulable.
        """
        return await self.async_uncordon()

    async def async_uncordon(self) -> None:
        await self.async_patch({"spec": {"unschedulable": False}})

    async def taint(self, key: str, value: str, *, effect: str) -> None:
        """Taint a node."""
        return await self.async_taint(key, value, effect=effect)

    async def async_taint(self, key: str, value: str, *, effect: str) -> None:
        await self.async_refresh()
        if effect.endswith("-"):
            # Remove taint with key
            effect = effect[:-1]
            if all(taint["key"] != key for taint in self.taints):
                raise NotFoundError(f"Unable to find taint with key: {key}")

            taints = [taint for taint in self.taints if taint["key"] != key]
        else:
            taints = list(self.taints) + [
                {"key": key, "value": value, "effect": effect}
            ]

        await self.async_patch({"spec": {"taints": taints}})

    @property
    def taints(self) -> Box:
        """Labels of the Kubernetes resource."""
        try:
            return self.raw["spec"]["taints"]
        except KeyError:
            return Box({})


class PersistentVolumeClaim(APIObject):
    """A Kubernetes PersistentVolumeClaim."""

    version = "v1"
    endpoint = "persistentvolumeclaims"
    kind = "PersistentVolumeClaim"
    plural = "persistentvolumeclaims"
    singular = "persistentvolumeclaim"
    namespaced = True


class PersistentVolume(APIObject):
    """A Kubernetes PersistentVolume."""

    version = "v1"
    endpoint = "persistentvolumes"
    kind = "PersistentVolume"
    plural = "persistentvolumes"
    singular = "persistentvolume"
    namespaced = False


class Pod(APIObject):
    """A Kubernetes Pod."""

    version = "v1"
    endpoint = "pods"
    kind = "Pod"
    plural = "pods"
    singular = "pod"
    namespaced = True

    async def async_ready(self) -> bool:
        """Check if the pod is ready."""
        await self.async_refresh()
        conditions = list_dict_unpack(
            self.status.get("conditions", []),
            key="type",
            value="status",
        )
        return (
            "Ready" in conditions
            and "ContainersReady" in conditions
            and conditions.get("Ready", "False") == "True"
            and conditions.get("ContainersReady", "False") == "True"
        )

    async def ready(self) -> bool:
        """Check if the pod is ready."""
        return await self.async_ready()

    async def logs(
        self,
        container=None,
        pretty=None,
        previous=False,
        since_seconds=None,
        since_time=None,
        timestamps=False,
        tail_lines=None,
        limit_bytes=None,
        follow=False,
        timeout=3600,
    ) -> AsyncGenerator[str]:
        async for line in self.async_logs(
            container=container,
            pretty=pretty,
            previous=previous,
            since_seconds=since_seconds,
            since_time=since_time,
            timestamps=timestamps,
            tail_lines=tail_lines,
            limit_bytes=limit_bytes,
            follow=follow,
            timeout=timeout,
        ):
            yield line

    async def async_logs(
        self,
        container=None,
        pretty=None,
        previous=False,
        since_seconds=None,
        since_time=None,
        timestamps=False,
        tail_lines=None,
        limit_bytes=None,
        follow=False,
        timeout=3600,
    ) -> AsyncGenerator[str]:
        """Streams logs from a Pod.

        Args:
            container:
                The container to get logs from. Defaults to the first container in the Pod.
            pretty:
                If True, return pretty logs. Defaults to False.
            previous:
                If True, return previous terminated container logs. Defaults to False.
            since_seconds:
                If set, return logs since this many seconds ago.
            since_time:
                If set, return logs since this time.
            timestamps:
                If True, prepend each log line with a timestamp. Defaults to False.
            tail_lines:
                If set, return this many lines from the end of the logs.
            limit_bytes:
                If set, return this many bytes from the end of the logs.
            follow:
                If True, follow the logs until the timeout is reached. Defaults to False.
            timeout:
                If following timeout after this many seconds. Set to None to disable timeout.

        Returns:
            An async generator yielding log lines.


        Example:
            >>> from kr8s.objects import Pod
            >>> pod = Pod.get("my-pod")
            >>> for line in pod.logs():
            ...     print(line)

            We can also follow logs as they are generated, the generator will yield a new log line as
            it is generated by the Pod. This blocks indefinitely so we can set a timeout to break
            after some period of time, the default is ``3600`` (1hr) but can be set to ``None`` to
            disable the timeout.

            >>> from kr8s.objects import Pod
            >>> pod = Pod.get("my-pod", namespace="ns")
            >>> for line in pod.logs(follow=True, timeout=60):
            ...     # Will continue streaming logs until 60 seconds or the Pod is terminated
            ...     print(line)

        """
        params: dict[Any, Any] = {}
        if follow:
            params["follow"] = "true"
        if container is not None:
            params["container"] = container
        if pretty is not None:
            params["pretty"] = pretty
        if previous:
            params["previous"] = "true"
        if since_seconds is not None and since_time is None:
            params["sinceSeconds"] = int(since_seconds)
        elif since_time is not None and since_seconds is None:
            params["sinceTime"] = since_time
        if timestamps:
            params["timestamps"] = "true"
        if tail_lines is not None:
            params["tailLines"] = int(tail_lines)
        if limit_bytes is not None:
            params["limitBytes"] = int(limit_bytes)

        with contextlib.suppress(httpx.ReadTimeout):
            assert self.api
            async with self.api.call_api(
                "GET",
                version=self.version,
                url=f"{self.endpoint}/{self.name}/log",
                namespace=self.namespace,
                params=params,
                stream=True,
                timeout=timeout,
            ) as resp:
                async for line in resp.aiter_lines():
                    yield line

    def portforward(
        self,
        remote_port: int,
        local_port: LocalPortType = "match",
        address: list[str] | str = "127.0.0.1",
    ) -> AsyncPortForward:
        """Port forward a pod.

        Returns an instance of :class:`kr8s.portforward.PortForward` for this Pod.

        Args:
            remote_port:
                The port on the Pod to forward to.
            local_port:
                The local port to listen on. Defaults to ``"match"``, which will match the ``remote_port``.
                Set to ``"auto"`` or ``None`` to find an available high port.
                Set to an ``int`` to specify a specific port.
            address:
                List of addresses or address to listen on. Defaults to ["127.0.0.1"], will listen only on 127.0.0.1.

        Example:
            This can be used as a an async context manager or with explicit start/stop methods.

            Context manager:

            >>> async with pod.portforward(8888) as port:
            ...     print(f"Forwarding to port {port}")
            ...     # Do something with port 8888


            Explict start/stop:

            >>> pf = pod.portforward(8888)
            >>> await pf.start()
            >>> print(f"Forwarding to port {pf.local_port}")
            >>> # Do something with port 8888
            >>> await pf.stop()

            Explict bind address:

            >>> async with pod.PortForward(8888, address=["127.0.0.1", "10.20.0.1"]) as port:
            ...     print(f"Forwarding to port {port}")
            ...     # Do something with port 8888 on the Pod, port will be bind to 127.0.0.1 and 10.20.0.1

        """
        if isinstance(address, str):
            address = [address]
        if self._asyncio:
            return AsyncPortForward(self, remote_port, local_port, address)
        return SyncPortForward(self, remote_port, local_port, address)

    async def async_exec(
        self,
        command: list[str],
        *,
        container: str | None = None,
        stdin: str | BinaryIO | None = None,
        stdout: BinaryIO | None = None,
        stderr: BinaryIO | None = None,
        check: bool = True,
        capture_output: bool = True,
    ):
        while not await self.async_ready():
            await anyio.sleep(0.1)

        ex = Exec(
            self,
            command,
            container=container,
            stdout=stdout,
            stderr=stderr,
            stdin=stdin,
            check=check,
            capture_output=capture_output,
        )
        async with ex.run() as process:
            await process.wait()
            return process.as_completed()

    async def exec(
        self,
        command: list[str],
        *,
        container: str | None = None,
        stdin: str | BinaryIO | None = None,
        stdout: BinaryIO | None = None,
        stderr: BinaryIO | None = None,
        check: bool = True,
        capture_output: bool = True,
    ):
        """Run a command in a container and wait until it completes.

        Behaves like :func:`subprocess.run`.

        Args:
            command:
                Command to execute.
            container:
                Container to execute the command in.
            stdin:
                If set, read stdin to the container.
            stdout:
                If set, write stdout to the provided writable stream object.
            stderr:
                If set, write stderr to the provided writable stream object.
            check:
                If True, raise an exception if the command fails.
            capture_output:
                If True, store stdout and stderr from the container in an attribute.

        Returns:
            A :class:`kr8s._exec.CompletedExec` object.

        Example:
            >>> from kr8s.objects import Pod
            >>> pod = Pod.get("my-pod")
            >>> ex = await pod.exec(["ls", "-l"])
            >>> print(ex.stdout)
            >>> print(ex.stderr)
        """
        return await self.async_exec(
            command,
            container=container,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            check=check,
            capture_output=capture_output,
        )

    @classmethod
    def gen(
        cls,
        *,
        name=None,
        generate_name=None,
        image=None,
        namespace=None,
        annotations=None,
        command=None,
        env=None,
        resources=None,
        image_pull_policy=None,
        labels=None,
        ports=None,
        restart="Always",
    ):
        """Generate a pod definition.

        Args:
            name (str): The name of the pod.
            generate_name (str): Template for generating the name of the pod.
            namespace (str): The namespace of the pod.
            image (str): The image to use.
            annotations (dict): Annotations to add to the pod.
            command (list): Command to run in the container.
            env (dict): Environment variables to set in the container.
            resources (dict): Resources to set in the container.
            image_pull_policy (str): Image pull policy to use.
            labels (dict): Labels to add to the pod.
            ports (list|int): Ports to expose.
            restart (str): Restart policy to use.

        Returns:
            A :class:`kr8s.objects.Pod` object.

        Example:
            >>> from kr8s.objects import Pod
            >>> pod = Pod.gen(name="my-pod", image="my-image")
            >>> pod.create()

            Create an nginx Pod that exposes port 80.
            >>> from kr8s.objects import Pod
            >>> pod = Pod.gen(name="nginx", image="nginx:latest", ports=[80])
            >>> pod.create()

            Create an wordpress Pod that exposes port 80.
            >>> from kr8s.objects import Pod
            >>> pod = Pod.gen(name="wordpress", image="wordpress:latest", ports=[{"containerPort": 80}])
            >>> pod.create()

            Create a Pod that requires a GPU
            >>> from kr8s.objects import Pod
            >>> pod = Pod.gen(name="cuda-vectoradd",
            ...               image="nvidia/samples:vectoradd-cuda11.6.0-ubuntu18.04",
            ...               resources={"limits": {"nvidia.com/gpu": 1}})
        """
        if not image:
            raise ValueError("Image must be specified")
        if not name and not generate_name:
            raise ValueError("Name or generate_name must be specified")
        if ports:
            if isinstance(ports, int):
                ports = [ports]
            # Convert any integer ports to valid v1.ContainerPort format
            ports = [
                {"containerPort": port} if isinstance(port, int) else port
                for port in ports
            ]

        return cls(
            xdict(
                apiVersion="v1",
                kind="Pod",
                metadata=xdict(
                    name=name,
                    generateName=generate_name,
                    namespace=namespace,
                    annotations=annotations,
                    labels=labels,
                ),
                spec=xdict(
                    containers=[
                        xdict(
                            name=name if name else "container",
                            image=image,
                            command=command,
                            env=env,
                            resources=resources,
                            imagePullPolicy=image_pull_policy,
                            ports=ports,
                        )
                    ],
                    restartPolicy=restart,
                ),
            )
        )

    async def tolerate(
        self,
        key: str,
        *,
        operator: str,
        effect: str,
        value: str | None = None,
        toleration_seconds: int | None = None,
    ):
        """Add a toleration to the Pod.

        Args:
            key (str): Key of the taint to tolerate.
            operator (str): Toleration operator.
            effect (str): Specifies the effect of the taint to tolerate.
            value (str): Value of taint to tolerate.
            toleration_seconds (str): Toleration seconds.
        """
        return await self.async_tolerate(
            key,
            operator=operator,
            effect=effect,
            value=value,
            toleration_seconds=toleration_seconds,
        )

    async def async_tolerate(
        self,
        key: str,
        *,
        operator: str,
        effect: str,
        value: str | None = None,
        toleration_seconds: int | None = None,
    ):
        new_toleration: dict = {"key": key, "operator": operator, "effect": effect}
        if value is not None:
            new_toleration["value"] = value

        if toleration_seconds is not None:
            new_toleration["tolerationSeconds"] = toleration_seconds

        self.tolerations.append(new_toleration)

        await self.async_patch({"spec": {"tolerations": self.tolerations}})

    @property
    def tolerations(self) -> Box:
        """Tolerations for Pod."""
        try:
            return self.raw["spec"]["tolerations"]
        except KeyError:
            return Box([])


class PodTemplate(APIObject):
    """A Kubernetes PodTemplate."""

    version = "v1"
    endpoint = "podtemplates"
    kind = "PodTemplate"
    plural = "podtemplates"
    singular = "podtemplate"
    namespaced = True


class ReplicationController(APIObject):
    """A Kubernetes ReplicationController."""

    version = "v1"
    endpoint = "replicationcontrollers"
    kind = "ReplicationController"
    plural = "replicationcontrollers"
    singular = "replicationcontroller"
    namespaced = True
    scalable = True

    async def ready(self):
        """Check if the deployment is ready."""
        return await self.async_ready()

    async def async_ready(self):
        await self.async_refresh()
        return (
            self.raw["status"].get("observedGeneration", 0)
            >= self.raw["metadata"]["generation"]
            and self.raw["status"].get("readyReplicas", 0) == self.replicas
        )


class ResourceQuota(APIObject):
    """A Kubernetes ResourceQuota."""

    version = "v1"
    endpoint = "resourcequotas"
    kind = "ResourceQuota"
    plural = "resourcequotas"
    singular = "resourcequota"
    namespaced = True


class Secret(APIObject):
    """A Kubernetes Secret."""

    version = "v1"
    endpoint = "secrets"
    kind = "Secret"
    plural = "secrets"
    singular = "secret"
    namespaced = True

    @property
    def data(self) -> Box:
        """Data of the Secret."""
        return self.raw["data"]

    @data.setter
    def data(self, value: dict) -> None:
        self.raw["data"] = value


class ServiceAccount(APIObject):
    """A Kubernetes ServiceAccount."""

    version = "v1"
    endpoint = "serviceaccounts"
    kind = "ServiceAccount"
    plural = "serviceaccounts"
    singular = "serviceaccount"
    namespaced = True


class Service(APIObject):
    """A Kubernetes Service."""

    version = "v1"
    endpoint = "services"
    kind = "Service"
    plural = "services"
    singular = "service"
    namespaced = True

    async def proxy_http_request(
        self, method: str, path: str, port: int | None = None, **kwargs: Any
    ) -> httpx.Response:
        """Issue a HTTP request with specific HTTP method to proxy of a Service.

        Args:
            method: HTTP method to use.
            path: Path to proxy.
            port: Port to proxy to. If not specified, the first port in the
                Service's spec will be used.
            **kwargs: Additional keyword arguments to pass to the API call.
        """
        return await self.async_proxy_http_request(method, path, port, **kwargs)

    async def async_proxy_http_request(
        self, method: str, path: str, port: int | None = None, **kwargs: Any
    ) -> httpx.Response:
        if port is None:
            port = self.raw["spec"]["ports"][0]["port"]
        assert self.api
        async with self.api.call_api(
            method,
            version=self.version,
            url=f"{self.endpoint}/{self.name}:{port}/proxy/{path}",
            namespace=self.namespace,
            **kwargs,
        ) as response:
            return response

    async def proxy_http_get(
        self, path: str, port: int | None = None, **kwargs
    ) -> httpx.Response:
        return await self.async_proxy_http_request("GET", path, port, **kwargs)

    async def proxy_http_post(
        self, path: str, port: int | None = None, **kwargs
    ) -> None:
        await self.async_proxy_http_request("POST", path, port, **kwargs)

    async def proxy_http_put(
        self, path: str, port: int | None = None, **kwargs
    ) -> httpx.Response:
        return await self.async_proxy_http_request("PUT", path, port, **kwargs)

    async def proxy_http_delete(
        self, path: str, port: int | None = None, **kwargs
    ) -> httpx.Response:
        return await self.async_proxy_http_request("DELETE", path, port, **kwargs)

    async def ready_pods(self) -> list[Pod]:
        """Return a list of ready Pods for this Service."""
        return await self.async_ready_pods()

    async def async_ready_pods(self) -> list[Pod]:
        """Return a list of ready Pods for this Service."""
        assert self.api
        pods = [
            pod
            async for pod in self.api.async_get(
                "pods",
                label_selector=dict_to_selector(self.spec["selector"]),
                namespace=self.namespace,
            )
        ]
        if isinstance(pods, Pod):
            pods = [pods]
        elif isinstance(pods, List) and all(isinstance(pod, Pod) for pod in pods):
            # The all(isinstance(...) for ...) check doesn't seem to narrow the type
            # correctly in pyright so we need to explicitly use cast
            pods = cast(List[Pod], pods)
        else:
            raise TypeError(f"Unexpected type {type(pods)} returned from API")
        return [pod for pod in pods if await pod.async_ready()]

    async def ready(self) -> bool:
        """Check if the service is ready."""
        return await self.async_ready()

    async def async_ready(self) -> bool:
        await self.async_refresh()

        # If the service is of type LoadBalancer, check if it has endpoints
        if (
            self.spec.type == "LoadBalancer"
            and len(self.status.loadBalancer.ingress or []) == 0
        ):
            return False

        # Check there is at least one Pod in service
        pods = await self.async_ready_pods()
        return len(pods) > 0

    def portforward(
        self,
        remote_port: int,
        local_port: LocalPortType = "match",
        address: str | list[str] = "127.0.0.1",
    ) -> AsyncPortForward:
        """Port forward a service.

        Returns an instance of :class:`kr8s.portforward.PortForward` for this Service.

        Args:
            remote_port:
                The port on the Pod to forward to.
            local_port:
                The local port to listen on. Defaults to ``"match"``, which will match the ``remote_port``.
                Set to ``"auto"`` or ``None`` to find an available high port.
                Set to an ``int`` to specify a specific port.
            address:
                List of addresses or address to listen on. Defaults to ["127.0.0.1"], will listen only on 127.0.0.1.

        Example:
            This can be used as a an async context manager or with explicit start/stop methods.

            Context manager:

            >>> async with service.portforward(8888) as port:
            ...     print(f"Forwarding to port {port}")
            ...     # Do something with port 8888


            Explict start/stop:

            >>> pf = service.portforward(8888)
            >>> await pf.start()
            >>> print(f"Forwarding to port {pf.local_port}")
            >>> # Do something with port 8888
            >>> await pf.stop()

        """
        if isinstance(address, str):
            address = [address]
        if self._asyncio:
            return AsyncPortForward(self, remote_port, local_port, address)
        return SyncPortForward(self, remote_port, local_port, address)


## apps/v1 objects


class ControllerRevision(APIObject):
    """A Kubernetes ControllerRevision."""

    version = "apps/v1"
    endpoint = "controllerrevisions"
    kind = "ControllerRevision"
    plural = "controllerrevisions"
    singular = "controllerrevision"
    namespaced = True


class DaemonSet(APIObject):
    """A Kubernetes DaemonSet."""

    version = "apps/v1"
    endpoint = "daemonsets"
    kind = "DaemonSet"
    plural = "daemonsets"
    singular = "daemonset"
    namespaced = True


class Deployment(APIObject):
    """A Kubernetes Deployment."""

    version = "apps/v1"
    endpoint = "deployments"
    kind = "Deployment"
    plural = "deployments"
    singular = "deployment"
    namespaced = True
    scalable = True

    async def pods(self) -> list[Pod]:
        """Return a list of Pods for this Deployment."""
        return await self.async_pods()

    async def async_pods(self) -> list[Pod]:
        assert self.api
        pods = [
            pod
            async for pod in self.api.async_get(
                "pods",
                label_selector=dict_to_selector(self.spec["selector"]["matchLabels"]),
                namespace=self.namespace,
            )
        ]
        if isinstance(pods, Pod):
            return [pods]
        if isinstance(pods, List) and all(isinstance(pod, Pod) for pod in pods):
            # The all(isinstance(...) for ...) check doesn't seem to narrow the type
            # correctly in pyright so we need to explicitly use cast
            return cast(List[Pod], pods)
        raise TypeError(f"Unexpected type {type(pods)} returned from API")

    async def ready(self):
        """Check if the deployment is ready."""
        return await self.async_ready()

    async def async_ready(self):
        await self.async_refresh()
        return (
            self.raw["status"].get("observedGeneration", 0)
            >= self.raw["metadata"]["generation"]
            and self.raw["status"].get("readyReplicas", 0) == self.replicas
        )


class ReplicaSet(APIObject):
    """A Kubernetes ReplicaSet."""

    version = "apps/v1"
    endpoint = "replicasets"
    kind = "ReplicaSet"
    plural = "replicasets"
    singular = "replicaset"
    namespaced = True
    scalable = True


class StatefulSet(APIObject):
    """A Kubernetes StatefulSet."""

    version = "apps/v1"
    endpoint = "statefulsets"
    kind = "StatefulSet"
    plural = "statefulsets"
    singular = "statefulset"
    namespaced = True
    scalable = True


## autoscaling/v1 objects


class HorizontalPodAutoscaler(APIObject):
    """A Kubernetes HorizontalPodAutoscaler."""

    version = "autoscaling/v2"
    endpoint = "horizontalpodautoscalers"
    kind = "HorizontalPodAutoscaler"
    plural = "horizontalpodautoscalers"
    singular = "horizontalpodautoscaler"
    namespaced = True


## batch/v1 objects


class CronJob(APIObject):
    """A Kubernetes CronJob."""

    version = "batch/v1"
    endpoint = "cronjobs"
    kind = "CronJob"
    plural = "cronjobs"
    singular = "cronjob"
    namespaced = True


class Job(APIObject):
    """A Kubernetes Job."""

    version = "batch/v1"
    endpoint = "jobs"
    kind = "Job"
    plural = "jobs"
    singular = "job"
    namespaced = True
    scalable = True
    scalable_spec = "parallelism"


## networking.k8s.io/v1 objects


class IngressClass(APIObject):
    """A Kubernetes IngressClass."""

    version = "networking.k8s.io/v1"
    endpoint = "ingressclasses"
    kind = "IngressClass"
    plural = "ingressclasses"
    singular = "ingressclass"
    namespaced = False


class Ingress(APIObject):
    """A Kubernetes Ingress."""

    version = "networking.k8s.io/v1"
    endpoint = "ingresses"
    kind = "Ingress"
    plural = "ingresses"
    singular = "ingress"
    namespaced = True


class NetworkPolicy(APIObject):
    """A Kubernetes NetworkPolicy."""

    version = "networking.k8s.io/v1"
    endpoint = "networkpolicies"
    kind = "NetworkPolicy"
    plural = "networkpolicies"
    singular = "networkpolicy"
    namespaced = True


## policy/v1 objects


class PodDisruptionBudget(APIObject):
    """A Kubernetes PodDisruptionBudget."""

    version = "policy/v1"
    endpoint = "poddisruptionbudgets"
    kind = "PodDisruptionBudget"
    plural = "poddisruptionbudgets"
    singular = "poddisruptionbudget"
    namespaced = True


## rbac.authorization.k8s.io/v1 objects


class ClusterRoleBinding(APIObject):
    """A Kubernetes ClusterRoleBinding."""

    version = "rbac.authorization.k8s.io/v1"
    endpoint = "clusterrolebindings"
    kind = "ClusterRoleBinding"
    plural = "clusterrolebindings"
    singular = "clusterrolebinding"
    namespaced = False


class ClusterRole(APIObject):
    """A Kubernetes ClusterRole."""

    version = "rbac.authorization.k8s.io/v1"
    endpoint = "clusterroles"
    kind = "ClusterRole"
    plural = "clusterroles"
    singular = "clusterrole"
    namespaced = False


class RoleBinding(APIObject):
    """A Kubernetes RoleBinding."""

    version = "rbac.authorization.k8s.io/v1"
    endpoint = "rolebindings"
    kind = "RoleBinding"
    plural = "rolebindings"
    singular = "rolebinding"
    namespaced = True


class Role(APIObject):
    """A Kubernetes Role."""

    version = "rbac.authorization.k8s.io/v1"
    endpoint = "roles"
    kind = "Role"
    plural = "roles"
    singular = "role"
    namespaced = True


## apiextensions.k8s.io/v1 objects


class CustomResourceDefinition(APIObject):
    """A Kubernetes CustomResourceDefinition."""

    version = "apiextensions.k8s.io/v1"
    endpoint = "customresourcedefinitions"
    kind = "CustomResourceDefinition"
    plural = "customresourcedefinitions"
    singular = "customresourcedefinition"
    namespaced = False


## meta.k8s.io/v1 objects


class Table(APIObject):
    """A Kubernetes Table."""

    version = "meta.k8s.io/v1"
    endpoint = "tables"
    kind = "Table"
    plural = "tables"
    singular = "table"
    namespaced = False

    @property
    def rows(self) -> list[dict]:
        """Table rows."""
        return self.raw["rows"]

    @property
    def column_definitions(self) -> list[dict]:
        """Table column definitions."""
        return self.raw["columnDefinitions"]


def get_class(
    kind: str,
    version: str | None = None,
    _asyncio: bool = True,
) -> type[APIObject]:
    """Get an APIObject subclass by kind and version.

    Args:
        kind: The Kubernetes resource kind.
        version: The Kubernetes API group/version.

    Returns:
        An APIObject subclass.

    Raises:
        KeyError: If no object is registered for the given kind and version.
    """
    result = None
    group = None
    if "/" in kind:
        kind, version = kind.split("/", 1)
    if "." in kind:
        kind, group = kind.split(".", 1)
    if version and "/" in version:
        if group:
            raise ValueError("Cannot specify group in both kind and version")
        group, version = version.split("/", 1)
    kind = kind.lower()

    def _walk_subclasses(cls):
        yield cls
        for subcls in cls.__subclasses__():
            yield from _walk_subclasses(subcls)

    for cls in _walk_subclasses(APIObject):
        if not hasattr(cls, "version"):
            continue
        if "/" in cls.version:
            cls_group, cls_version = cls.version.split("/")
        else:
            cls_group, cls_version = None, cls.version
        if (
            hasattr(cls, "kind")
            and cls._asyncio == _asyncio
            and (cls.kind == kind or cls.singular == kind or cls.plural == kind)
        ):
            if (group is None or cls_group == group) and (
                version is None or cls_version == version
            ):
                result = cls
            if (
                group
                and not version
                and "." in group
                and cls_group == group.split(".", 1)[1]
                and cls_version == group.split(".", 1)[0]
            ):
                result = cls

    if result:
        return result
    raise KeyError(
        f"No object registered for {kind}{'.' + group if group else ''}. "
        "See https://docs.kr8s.org/en/stable/object.html#extending-the-objects-api "
        "for more information on how to register a new object."
    )


def new_class(
    kind: str,
    version: str | None = None,
    asyncio: bool = True,
    namespaced=True,
    scalable: bool | None = None,
    scalable_spec: str | None = None,
    plural: str | None = None,
) -> type[APIObject]:
    """Create a new APIObject subclass.

    Args:
        kind: The Kubernetes resource kind.
        version: The Kubernetes API version.
        asyncio: Whether to use asyncio or not.
        namespaced: Whether the resource is namespaced or not.
        scalable: Whether the resource is scalable or not.
        scalable_spec: The name of the field to use for scaling.
        plural: The plural form of the resource.

    Returns:
        A new APIObject subclass.
    """
    if "." in kind:
        kind, version = kind.split(".", 1)
    if version is None:
        version = "v1"
    plural = plural or kind.lower() + "s"
    construct_args = {
        "kind": kind,
        "version": version,
        "_asyncio": asyncio,
        "endpoint": plural.lower(),
        "plural": plural.lower(),
        "singular": kind.lower(),
        "namespaced": namespaced,
        "scalable": scalable or False,
        "scalable_spec": scalable_spec or "replicas",
    }
    if asyncio:
        return type(kind, (APIObject,), construct_args)
    else:
        return type(
            kind,
            (
                APIObjectSyncMixin,
                APIObject,
            ),
            construct_args,
        )


def object_from_spec(
    spec: dict,
    api: Api | None = None,
    allow_unknown_type: bool = False,
    _asyncio: bool = True,
) -> APIObject:
    """Create an APIObject from a Kubernetes resource spec.

    Args:
        spec: A Kubernetes resource spec.
        api: An optional API instance to use.
        allow_unknown_type: Whether to allow unknown resource types.
        _asyncio: Whether to use asyncio or not.

    Returns:
        A corresponding APIObject subclass instance.

    Raises:
        ValueError: If the resource kind or API version is not supported.
    """
    try:
        cls = get_class(spec["kind"], spec["apiVersion"], _asyncio=_asyncio)
    except KeyError:
        if allow_unknown_type:
            cls = new_class(spec["kind"], spec["apiVersion"], asyncio=_asyncio)
        else:
            raise
    return cls(spec, api=api)


async def object_from_name_type(
    name: str,
    namespace: str | None = None,
    api: Api | None = None,
    _asyncio: bool = True,
) -> APIObject:
    """Create an APIObject from a Kubernetes resource name.

    Args:
        name: A Kubernetes resource name.
        namespace: The namespace of the resource.
        api: An optional API instance to use.
        _asyncio: Whether to use asyncio or not.

    Returns:
        A corresponding APIObject subclass instance.

    Raises:
        ValueError: If the resource kind or API version is not supported.
    """
    if "/" not in name:
        raise ValueError(f"Invalid name: {name}. Expecting format of 'resource/name'.")
    resource_type, name = name.split("/")
    if "." in resource_type:
        kind = resource_type.split(".")[0]
        version = ".".join(resource_type.split(".")[1:])
    else:
        kind = resource_type
        version = None
    cls = get_class(kind, version, _asyncio=_asyncio)
    return await cls.get(name, namespace=namespace, api=api)


async def objects_from_files(
    path: str | pathlib.Path,
    api: Api | None = None,
    recursive: bool = False,
    _asyncio: bool = True,
) -> list[APIObject]:
    """Create APIObjects from Kubernetes resource files.

    Args:
        path: A path to a Kubernetes resource file or directory of resource files.
        api: An optional API instance to use.
        recursive: Whether to recursively search for resource files in subdirectories.
        _asyncio: Whether to use asyncio or not.

    Returns:
        A list of APIObject subclass instances.

    Raises:
        ValueError: If the resource kind or API version is not supported.
    """
    path = pathlib.Path(path)
    if path.is_dir():
        pattern = "**/*.yaml" if recursive else "*.yaml"
        files = [f for f in path.glob(pattern) if f.is_file()]
    else:
        files = [path]
    if not api:
        api = await kr8s.asyncio.api(_asyncio=_asyncio)
    objects = []
    for file in files:
        with open(file) as f:
            for doc in yaml.safe_load_all(f):
                if doc is not None:
                    obj = object_from_spec(
                        doc, api=api, allow_unknown_type=True, _asyncio=_asyncio
                    )
                    if _asyncio:
                        await obj
                    objects.append(obj)
    return objects


def parse_kind(kind: str) -> tuple[str, str, str]:
    """Parse a Kubernetes resource kind into a tuple of (kind, group, version).

    Args:
        kind: The Kubernetes resource kind.

    Returns:
        A tuple of (kind, group, version).

    Example:
        >>> parse_kind("Pod")
        ("pod", "", "")
        >>> parse_kind("deployment")
        ("deployment", "", "")
        >>> parse_kind("services/v1")
        ("services", "", "v1")
        >>> parse_kind("ingress.networking.k8s.io/v1")
        ("ingress", "networking.k8s.io", "v1")
        >>> parse_kind("role.v1.rbac.authorization.k8s.io")
        ("role", "rbac.authorization.k8s.io", "v1")
    """
    if "/" in kind:
        kind, version = kind.split("/", 1)
    else:
        version = ""
    if "." in kind:
        kind, group = kind.split(".", 1)
    else:
        group = ""
    if "." in group:
        first, potential_group = group.split(".", 1)
        if re.match(r"v\d[a-z0-9]*", first):
            version = first
            group = potential_group
    return kind.lower(), group.lower(), version.lower()



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
#
# Utilities for running async code in a sync context. This is how kr8s is able to provide a sync API.
#
# The sync API in kr8s needs to wrap all async functions in a way that allows them to be run in a sync context.
# However, the sync API may be used within an async context, so we need to be able to run async code nested in
# a sync context that is itself nested in an async context. This is a bit of a tricky problem, but it can be
# solved by running the nested async code in a separate thread.
#
# This file was originally based on universalasync (commit d397911) and jupyter-core (commit 98b9a1a).
# Both projects attempt to solve the same problem: how to run nested async tasks.
# Neither solution quite fit in here, so we forked them and combined them. Things have evolved a lot
# since then, but the licenses and links are included here for historical reasons.
#
# universalasync License: https://github.com/bitcartcc/universalasync/blob/d397911/LICENSE
# jupyter-core License: https://github.com/jupyter/jupyter_core/blob/98b9a1a/COPYING.md
#
# This implementation now uses anyio to simplify dispatching to a loop in a thread using either
# asyncio or trio.
from __future__ import annotations

import builtins
import inspect
import subprocess
import sys
import tempfile
from contextlib import asynccontextmanager
from functools import partial, wraps
from threading import Thread
from typing import (
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Generator,
    TypeVar,
)

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

import anyio
import anyio.from_thread
import anyio.to_thread

T = TypeVar("T")
C = TypeVar("C")
P = ParamSpec("P")


class Portal:
    """A class that manages a thread running an anyio loop.

    This class is a singleton that manages a thread running an anyio loop and provides
    an anyio portal to communicate with the loop from a sync context.

    See https://anyio.readthedocs.io/en/stable/api.html#anyio.from_thread.start_blocking_portal for more info.

    It's important to start the loop in a separate thread because the sync code may be
    running in a context where an event loop is already running, and we can't run two
    event loops in the same thread. This commonly happens when running in IPython, a Jupyter
    notebook or when running async tests with pytest.

    """

    _instance: Portal
    _portal: anyio.from_thread.BlockingPortal
    thread: Thread

    def __new__(cls):
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)
            cls._instance.thread = Thread(
                target=anyio.run, args=[cls._instance._run], name="Kr8sSyncRunnerThread"
            )
            cls._instance.thread.daemon = True
            cls._instance.thread.start()
        return cls._instance

    async def _run(self):
        async with anyio.from_thread.BlockingPortal() as portal:
            self._portal = portal
            await portal.sleep_until_stopped()

    def call(self, func: Callable[P, Awaitable[T]], *args, **kwargs) -> T:
        """Call a coroutine in the runner loop and return the result."""
        # On first call the thread has to start the loop, so we need to wait for it
        while not hasattr(self, "_portal"):
            pass
        return self._portal.call(func, *args, **kwargs)


def run_sync(
    coro: Callable[P, AsyncGenerator | Awaitable[T]]
) -> Callable[P, Generator | T]:
    """Wraps a coroutine in a function that blocks until it has executed.

    Args:
        coro (Awaitable): A coroutine.

    Returns:
        Callable: A sync function that executes the coroutine via the :class`Portal`.
    """
    if inspect.isasyncgenfunction(coro):

        @wraps(coro)
        def run_gen_inner(*args: P.args, **kwargs: P.kwargs) -> Generator:
            return iter_over_async(coro(*args, **kwargs))

        return run_gen_inner

    if inspect.iscoroutinefunction(coro):

        @wraps(coro)
        def run_sync_inner(*args: P.args, **kwargs: P.kwargs) -> T:
            wrapped = partial(coro, *args, **kwargs)
            portal = Portal()
            return portal.call(wrapped)

        return run_sync_inner

    raise TypeError(f"Expected coroutine function, got {coro.__class__.__name__}")


def iter_over_async(agen: AsyncGenerator) -> Generator:
    """Convert an async generator to a sync generator.

    Args:
        agen (AsyncGenerator): async generator to convert

    Yields:
        Any: object from async generator
    """
    ait = agen.__aiter__()

    async def get_next() -> tuple[bool, Any]:
        try:
            obj = await ait.__anext__()
            return False, obj
        except StopAsyncIteration:
            return True, None

    portal = Portal()
    while True:
        done, obj = portal.call(get_next)
        if done:
            break
        yield obj


def sync(source: C) -> C:
    """Convert all public async methods/properties of an object to universal methods.

    Private methods or methods starting with "async_" are ignored.
    See :func:`run_sync` for more info on how the conversion works.

    Args:
        source (C): object with coroutines to convert

    Returns:
        C: converted object with sync methods

    Examples:
        It's common to implement a coroutine and name it with async_ and then wrap that
        in another corotine that calls it. That way the method can be called from both
        sync and async contexts as function allows you to convert the outer coroutine
        to a sync method. But other async methods or other async objects can call the
        inner coroutine directly.

        >>> class Foo:
        ...     async def async_bar(self):
        ...         return 42
        ...     async def bar(self):
        ...         return await self.async_bar()
        ...     async def baz(self):
        ...         # If you want to calll self.bar() from another async method
        ...         # you can't when it gets wrapped with sync, so you can call
        ...         # self.async_bar() directly instead.
        ...         return (await self.async_bar()) + 1
        ...
        >>> SyncFoo = sync(Foo)
        >>> inst = SyncFoo()
        >>> inst.bar()
        42
        >>> inst.async_bar()
        <coroutine object Foo.async_bar at 0x7fbe0442b940>
        >>> inst.baz()
        43

    """
    setattr(source, "_asyncio", False)  # noqa: B010
    for name in dir(source):
        method = getattr(source, name)

        if not name.startswith("_") and not name.startswith("async_"):
            if inspect.iscoroutinefunction(method) or inspect.isasyncgenfunction(
                method
            ):
                function = getattr(source, name)
                setattr(source, name, run_sync(function))

        elif name == "__aenter__" and not hasattr(source, "__enter__"):
            setattr(source, "__enter__", run_sync(method))  # noqa: B010

        elif name == "__aexit__" and not hasattr(source, "__exit__"):
            setattr(source, "__exit__", run_sync(method))  # noqa: B010

    return source


async def check_output(*args, **kwargs) -> str:
    """Run a command and return its output."""
    completed_process = await anyio.run_process(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=False,
        **kwargs,
    )
    if completed_process.returncode == 0:
        return completed_process.stdout.decode()
    else:
        raise RuntimeError(
            f"Process exited with non-zero code {completed_process.returncode}:\n{completed_process.stderr.decode()}"
        )


@asynccontextmanager
async def NamedTemporaryFile(  # noqa: N802
    *args, delete: bool = True, **kwargs
) -> AsyncGenerator[anyio.Path]:
    """Create a temporary file that is deleted when the context exits."""
    kwargs.update(delete=False)

    def f():
        return tempfile.NamedTemporaryFile(*args, **kwargs)

    tmp = await anyio.to_thread.run_sync(f)
    fh = anyio.Path(tmp.name)
    yield fh
    if delete:
        await fh.unlink()


# If Python is less than 3.10 we need to define anext
if sys.version_info < (3, 10):

    async def anext(ait: AsyncGenerator) -> Any:
        """Get the next item from an async generator."""
        return await ait.__anext__()

else:
    anext = builtins.anext



# SPDX-FileCopyrightText: Copyright (c) 2023-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import base64
import os
import socket
import subprocess
import tempfile
import uuid
from contextlib import closing
from pathlib import Path

import anyio
import pytest
import yaml

from kr8s._api import Api
from kr8s._testutils import set_env

HERE = Path(__file__).parent.resolve()
DEFAULT_LABELS = {"created-by": "kr8s-tests"}


@pytest.fixture
async def example_pod_spec(ns):
    name = "example-" + uuid.uuid4().hex[:10]
    return {
        "apiVersion": "v1",
        "kind": "Pod",
        "metadata": {
            "name": name,
            "namespace": ns,
            "labels": {"hello": "world", **DEFAULT_LABELS},
            "annotations": {"foo": "bar"},
        },
        "spec": {
            "containers": [{"name": "pause", "image": "gcr.io/google_containers/pause"}]
        },
    }


@pytest.fixture
async def bad_pod_spec(ns):
    name = "example-" + uuid.uuid4().hex[:10]
    return {
        "apiVersion": "v1",
        "kind": "Pod",
        "metadata": {
            "name": name,
            "namespace": ns,
        },
        "spec": {
            "containers": [
                {
                    "name1": "pause",  # This is bad
                    "image": "gcr.io/google_containers/pause",
                }
            ]
        },
    }


@pytest.fixture
async def example_service_spec(ns):
    name = "example-" + uuid.uuid4().hex[:10]
    return {
        "apiVersion": "v1",
        "kind": "Service",
        "metadata": {
            "name": name,
            "namespace": ns,
            "labels": {"hello": "world", **DEFAULT_LABELS},
            "annotations": {"foo": "bar"},
        },
        "spec": {
            "ports": [{"port": 80, "targetPort": 9376}],
            "selector": {"app": "MyApp"},
        },
    }


@pytest.fixture
async def example_deployment_spec(ns):
    name = "example-" + uuid.uuid4().hex[:10]
    return {
        "apiVersion": "apps/v1",
        "kind": "Deployment",
        "metadata": {
            "name": name,
            "namespace": ns,
            "labels": {"hello": "world", **DEFAULT_LABELS},
            "annotations": {"foo": "bar"},
        },
        "spec": {
            "replicas": 1,
            "selector": {"matchLabels": {"app": name}},
            "template": {
                "metadata": {"labels": {"app": name}},
                "spec": {
                    "containers": [
                        {
                            "name": "pause",
                            "image": "gcr.io/google_containers/pause",
                        }
                    ]
                },
            },
        },
    }


@pytest.fixture
async def example_crd_spec():
    spec = {
        "apiVersion": "apiextensions.k8s.io/v1",
        "kind": "CustomResourceDefinition",
        "metadata": {"name": "shirts.stable.example.com"},
        "spec": {
            "group": "stable.example.com",
            "scope": "Namespaced",
            "names": {"plural": "shirts", "singular": "shirt", "kind": "Shirt"},
            "versions": [
                {
                    "name": "v1",
                    "served": True,
                    "storage": True,
                    "schema": {
                        "openAPIV3Schema": {
                            "type": "object",
                            "properties": {
                                "spec": {
                                    "type": "object",
                                    "properties": {
                                        "color": {"type": "string"},
                                        "size": {"type": "string"},
                                    },
                                }
                            },
                        }
                    },
                    "selectableFields": [
                        {"jsonPath": ".spec.color"},
                        {"jsonPath": ".spec.size"},
                    ],
                    "additionalPrinterColumns": [
                        {"jsonPath": ".spec.color", "name": "Color", "type": "string"},
                        {"jsonPath": ".spec.size", "name": "Size", "type": "string"},
                    ],
                }
            ],
        },
    }
    yield spec


def check_socket(host, port):
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        return sock.connect_ex((host, port)) == 0


@pytest.fixture(scope="session", autouse=True)
def run_id():
    return uuid.uuid4().hex[:4]


@pytest.fixture(autouse=True)
def ns(k8s_cluster, run_id):
    name = f"kr8s-pytest-{run_id}-{uuid.uuid4().hex[:4]}"
    k8s_cluster.kubectl("create", "namespace", name)
    yield name
    k8s_cluster.kubectl("delete", "namespace", name, "--wait=false")


@pytest.fixture
async def kubectl_proxy(k8s_cluster):
    proxy = subprocess.Popen(
        [k8s_cluster.kubectl_path, "proxy"],
        env={**os.environ, "KUBECONFIG": str(k8s_cluster.kubeconfig_path)},
    )
    host = "localhost"
    port = 8001
    while not check_socket(host, port):
        await anyio.sleep(0.1)
    yield f"http://{host}:{port}"
    proxy.kill()


@pytest.fixture(scope="session")
def k8s_token(k8s_cluster):
    # Apply the serviceaccount.yaml
    k8s_cluster.kubectl(
        "apply", "-f", str(HERE / "tests" / "resources" / "serviceaccount.yaml")
    )
    yield k8s_cluster.kubectl("create", "token", "pytest")
    # Delete the serviceaccount.yaml
    k8s_cluster.kubectl(
        "delete", "-f", str(HERE / "tests" / "resources" / "serviceaccount.yaml")
    )


@pytest.fixture
def serviceaccount(k8s_cluster, k8s_token):
    # Load kubeconfig
    kubeconfig = yaml.safe_load(k8s_cluster.kubeconfig_path.read_text())

    # Get the server host and port from the kubeconfig
    _hostport = kubeconfig["clusters"][0]["cluster"]["server"].split("//")[1]
    host, port = _hostport.split(":")

    # Create a temporary directory and populate it with the serviceaccount files
    with (
        tempfile.TemporaryDirectory() as tempdir,
        set_env(KUBERNETES_SERVICE_HOST=host, KUBERNETES_SERVICE_PORT=port),
    ):
        tempdir = Path(tempdir)
        # Create ca.crt in tempdir from the certificate-authority-data in kubeconfig
        (tempdir / "ca.crt").write_text(
            base64.b64decode(
                kubeconfig["clusters"][0]["cluster"]["certificate-authority-data"]
            ).decode()
        )
        namespace = "default"
        (tempdir / "token").write_text(k8s_token)
        (tempdir / "namespace").write_text(namespace)
        yield str(tempdir)


@pytest.fixture(autouse=True)
def ensure_new_api_between_tests():
    yield
    Api._instances.clear()



# SPDX-FileCopyrightText: Copyright (c) 2024-2025, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
import pathlib
import typing
from typing import Any, Dict, List, Optional, Protocol, Union

import anyio
import jsonpath
import yaml

from kr8s._data_utils import dict_list_pack, list_dict_unpack
from kr8s._types import PathType

# TODO Implement set cluster
# TODO Implement delete cluster
# TODO Implement set context
# TODO Implement delete context
# TODO Implement set user
# TODO Implement delete user


class KubeConfigProtocol(Protocol):
    @property
    def raw(self) -> dict: ...


class KubeConfigMixin:

    def get(
        self: KubeConfigProtocol,
        path: Optional[str] = None,
        pointer: Optional[str] = None,
    ) -> Any:
        """Get a value from the config using a JSON Path or JSON Pointer."""
        if not path and not pointer:
            raise ValueError("No path or pointer provided")
        if path:
            return jsonpath.findall(path, self.raw)
        if pointer:
            return jsonpath.pointer.resolve(pointer, self.raw)


class KubeConfigSet(KubeConfigMixin):
    def __init__(self, *paths_or_dicts: Union[PathType, Dict]):
        self._configs = []
        for path_or_dict in paths_or_dicts:
            try:
                self._configs.append(KubeConfig(path_or_dict))
            except ValueError:
                pass
        if not self._configs:
            raise ValueError("No valid kubeconfig provided")

    def __await__(self):
        async def f():
            for config in self._configs:
                await config
            return self

        return f().__await__()

    async def save(self):
        for config in self._configs:
            await config.save()

    @property
    def path(self) -> PathType:
        return self.get_path()

    def get_path(self, context: Optional[str] = None) -> PathType:
        """Return the path of the config for the current context.

        Args:
            context (str): Override the context to use. If not provided, the current context is used.
        """
        if not context:
            context = self.current_context
        if context:
            for config in self._configs:
                if context in [c["name"] for c in config.contexts]:
                    return config.path
        return self._configs[0].path

    @property
    def raw(self) -> Dict:
        """Merge all kubeconfig data into a single kubeconfig."""
        data = {
            "apiVersion": "v1",
            "kind": "Config",
            "preferences": self.preferences,
            "clusters": self.clusters,
            "users": self.users,
            "contexts": self.contexts,
            "current-context": self.current_context,
        }
        if self.extensions:
            data["extensions"] = self.extensions
        return data

    @property
    def current_context(self) -> str:
        """Return the current context from the first kubeconfig.

        Context configuration from multiples files are ignored.
        """
        return self._configs[0].current_context

    @property
    def current_namespace(self) -> str:
        """Return the current namespace from the current context."""
        return self.get_context(self.current_context).get("namespace", "default")

    async def use_namespace(self, namespace: str) -> None:
        for config in self._configs:
            for context in config._raw["contexts"]:
                if context["name"] == self.current_context:
                    context["context"]["namespace"] = namespace
            await config.save()

    async def use_context(self, context: str) -> None:
        """Set the current context."""
        if context not in [c["name"] for c in self.contexts]:
            raise ValueError(f"Context {context} not found")
        await self._configs[0].use_context(context, allow_unknown=True)

    async def rename_context(self, old: str, new: str) -> None:
        """Rename a context."""
        for config in self._configs:
            if old in [c["name"] for c in config.contexts]:
                await config.rename_context(old, new)
                if self.current_context == old:
                    await self.use_context(new)
                return
        raise ValueError(f"Context {old} not found")

    def get_context(self, context_name: str) -> Dict:
        """Get a context by name."""
        for context in self.contexts:
            if context["name"] == context_name:
                return context["context"]
        raise ValueError(f"Context {context_name} not found")

    def get_cluster(self, cluster_name: str) -> Dict:
        """Get a cluster by name."""
        for cluster in self.clusters:
            if cluster["name"] == cluster_name:
                return cluster["cluster"]
        raise ValueError(f"Cluster {cluster_name} not found")

    def get_user(self, user_name: str) -> Dict:
        """Get a user by name."""
        for user in self.users:
            if user["name"] == user_name:
                return user["user"]
        raise ValueError(f"User {user_name} not found")

    async def set(self, pointer: str, value) -> None:
        """Replace a value using a JSON Pointer.

        Set only applies to the first kubeconfig.
        """
        await self._configs[0].set(pointer=pointer, value=value)
        await anyio.sleep(0)

    async def unset(self, pointer: str) -> None:
        """Remove a value using a JSON Pointer.

        Unset applies to all kubeconfigs.
        """
        for config in self._configs:
            try:
                await config.unset(pointer=pointer)
            except Exception:
                pass

    @property
    def preferences(self) -> List[Dict]:
        return self._configs[0].preferences

    @property
    def clusters(self) -> List[Dict]:
        clusters = []
        for config in self._configs:
            if config.clusters:
                clusters.extend(config.clusters)
        # Unpack and repack to remove duplicates
        unpacked = list_dict_unpack(clusters, "name", "cluster")
        repacked = dict_list_pack(unpacked, "name", "cluster")
        return repacked

    @property
    def users(self) -> List[Dict]:
        users = []
        for config in self._configs:
            if config.users:
                users.extend(config.users)
        # Unpack and repack to remove duplicates
        unpacked = list_dict_unpack(users, "name", "user")
        repacked = dict_list_pack(unpacked, "name", "user")
        return repacked

    @property
    def contexts(self) -> List[Dict]:
        contexts = []
        for config in self._configs:
            if config.contexts:
                contexts.extend(config.contexts)
        # Unpack and repack to remove duplicates
        unpacked = list_dict_unpack(contexts, "name", "context")
        repacked = dict_list_pack(unpacked, "name", "context")
        return repacked

    @property
    def extensions(self) -> List[Dict]:
        extensions = []
        for config in self._configs:
            if config.extensions:
                extensions.extend(config.extensions)
        return extensions


class KubeConfig(KubeConfigMixin):
    def __init__(self, path_or_config: Union[PathType, Dict]):
        self.path: PathType
        self._raw: dict = {}

        if not path_or_config:
            raise ValueError("KubeConfig path_or_config is None or empty string.")
        if isinstance(path_or_config, str) or isinstance(path_or_config, pathlib.Path):
            self.path = pathlib.Path(path_or_config).expanduser()
            if not self.path.exists():
                raise ValueError(f"File {self.path} does not exist")
            if self.path.is_dir():
                raise IsADirectoryError(
                    f'Error loading config file "{self.path}": is a directory.'
                )
        elif isinstance(path_or_config, dict):
            self._raw = path_or_config
        else:
            raise TypeError("KubeConfig path_or_config must be a string, path or dict.")

        self.__write_lock = anyio.Lock()

    def __await__(self):
        async def f():
            if not self._raw:
                async with await anyio.open_file(self.path) as fh:
                    self._raw = yaml.safe_load(await fh.read())
            return self

        return f().__await__()

    async def save(self, path=None) -> None:
        path = self.path if not path else path
        if not path:
            raise ValueError("No path provided")
        async with self.__write_lock:
            async with await anyio.open_file(path, "w") as fh:
                await fh.write(yaml.safe_dump(self._raw))

    @property
    def current_context(self) -> str:
        return self._raw["current-context"]

    @property
    def current_namespace(self) -> str:
        return self.get_context(self.current_context).get("namespace", "default")

    async def use_namespace(self, namespace: str) -> None:
        for context in self._raw["contexts"]:
            if context["name"] == self.current_context:
                context["context"]["namespace"] = namespace
        await self.save()

    async def use_context(self, context: str, allow_unknown: bool = False) -> None:
        """Set the current context."""
        if not allow_unknown and context not in [c["name"] for c in self.contexts]:
            raise ValueError(f"Context {context} not found")
        self._raw["current-context"] = context
        await self.save()

    async def rename_context(self, old: str, new: str) -> None:
        """Rename a context."""
        for context in self._raw["contexts"]:
            if context["name"] == old:
                context["name"] = new
                if self.current_context == old:
                    await self.use_context(new)
                await self.save()
                return
        raise ValueError(f"Context {old} not found")

    def get_context(self, context_name: str) -> Dict:
        """Get a context by name."""
        for context in self.contexts:
            if context["name"] == context_name:
                return context["context"]
        raise ValueError(f"Context {context_name} not found")

    def get_cluster(self, cluster_name: str) -> Dict:
        """Get a cluster by name."""
        for cluster in self.clusters:
            if cluster["name"] == cluster_name:
                return cluster["cluster"]
        raise ValueError(f"Cluster {cluster_name} not found")

    def get_user(self, user_name: str) -> Dict:
        """Get a user by name."""
        for user in self.users:
            if user["name"] == user_name:
                return user["user"]
        raise ValueError(f"User {user_name} not found")

    async def set(
        self, pointer: str, value: Optional[Any] = None, strict: bool = False
    ) -> None:
        """Replace a value using a JSON Pointer."""
        if strict:
            patch = jsonpath.JSONPatch().replace(pointer, value)
        else:
            patch = jsonpath.JSONPatch().add(pointer, value)
        self._raw = typing.cast(dict, patch.apply(self._raw))
        await self.save()

    async def unset(self, pointer: str) -> Any:
        """Remove a value using a JSON Pointer."""
        patch = jsonpath.JSONPatch().remove(pointer)
        self._raw = typing.cast(dict, patch.apply(self._raw))
        await self.save()

    @property
    def raw(self) -> Dict:
        return self._raw

    @property
    def preferences(self) -> List[Dict]:
        return self._raw["preferences"]

    @property
    def clusters(self) -> List[Dict]:
        return self._raw["clusters"]

    @property
    def users(self) -> List[Dict]:
        return self._raw["users"]

    @property
    def contexts(self) -> List[Dict]:
        return self._raw["contexts"]

    @property
    def extensions(self) -> List[Dict]:
        return self._raw["extensions"] if "extensions" in self._raw else []
